@@ -2,15 +2,14 @@ import collections.abc
 
 from zipfile import ZipFile, BadZipFile
 from pathlib import Path
-from typing import Sequence, overload, Union, Iterable, Iterator, Any, Optional, Tuple, Callable, Mapping, Type
-from coba.backports import Literal
+from typing import Sequence, overload, Union, Iterable, Iterator, Any, Optional, Tuple, Callable, Mapping, Type, Literal
 
 from coba                 import pipes
-from coba.contexts        import CobaContext, DiskCacher, DecoratedLogger, ExceptLog, NameLog, StampLog
+from coba.context         import CobaContext, DiskCacher, DecoratedLogger, ExceptLog, NameLog, StampLog
 from coba.pipes.sources   import DataFrameSource
 from coba.primitives      import Context, Action
 from coba.random          import CobaRandom
-from coba.pipes           import Pipes, Source, HttpSource, IterableSource, JsonDecode, Foreach
+from coba.pipes           import Pipes, Source, HttpSource, IterableSource, JsonDecode
 from coba.exceptions      import CobaException
 from coba.multiprocessing import CobaMultiprocessor
 from coba.learners        import Learner
@@ -23,11 +22,10 @@ from coba.environments.synthetics import LinearSyntheticSimulation, NeighborsSyn
 from coba.environments.synthetics import KernelSyntheticSimulation, MLPSyntheticSimulation, LambdaSimulation
 from coba.environments.supervised import SupervisedSimulation
 
-
 from coba.environments.filters   import EnvironmentFilter, Repr, Batch, Chunk, Logged, Finalize, BatchSafe
-from coba.environments.filters   import Binary, Shuffle, Take, Sparse, Reservoir, Cycle, Scale, Unbatch, Slice
-from coba.environments.filters   import Impute, Where, Noise, Riffle, Sort, Flatten, Cache, Params, Grounded
-from coba.environments.filters   import MappingToInteraction
+from coba.environments.filters   import Binary, Shuffle, Take, Sparsify, Densify, Reservoir, Cycle, Scale, Unbatch
+from coba.environments.filters   import Slice, Impute, Where, Noise, Riffle, Sort, Flatten, Cache, Params, Grounded
+from coba.environments.filters   import MappingToInteraction, OpeRewards
 
 from coba.environments.serialized import EnvironmentFromObjects, EnvironmentsToObjects, ZipMemberToObjects, ObjectsToZipMember
 
@@ -41,11 +39,11 @@ class Environments(collections.abc.Sequence, Sequence[Environment]):
 
     @overload
     @staticmethod
-    def from_template(filesource:Source[Iterable[str]]) -> 'Environments': ...
+    def from_template(filesource:Source[Iterable[str]], **user_vars) -> 'Environments': ...
 
     @overload
     @staticmethod
-    def from_template(fileurl:str) -> 'Environments': ...
+    def from_template(fileurl:str, **user_vars) -> 'Environments': ...
 
     @staticmethod
     def from_template(arg, **user_vars) -> 'Environments':
@@ -163,7 +161,8 @@ class Environments(collections.abc.Sequence, Sequence[Environment]):
         drop_missing: bool = True,
         take: int = None,
         *,
-        target:str = None) -> 'Environments':
+        target:str = None,
+        label_type:Literal['c','r','m'] = None) -> 'Environments':
         ...
 
     @overload
@@ -171,7 +170,8 @@ class Environments(collections.abc.Sequence, Sequence[Environment]):
     def from_openml(*,task_id: Union[int,Sequence[int]],
         drop_missing: bool = True,
         take: int = None,
-        target:str = None) -> 'Environments':
+        target:str = None,
+        label_type:Literal['m','c','r'] = None) -> 'Environments':
         ...
 
     @staticmethod
@@ -217,8 +217,10 @@ class Environments(collections.abc.Sequence, Sequence[Environment]):
 
     @staticmethod
     def from_save(path:str) -> 'Environments':
-        make = lambda m: EnvironmentFromObjects(ZipMemberToObjects(path,m))
-        return Environments(list(map(make,ZipFile(path).namelist())))
+        envs = []
+        for name in ZipFile(path).namelist():
+            envs.append(EnvironmentFromObjects(ZipMemberToObjects(path,name)))
+        return Environments(envs)
 
     @overload
     def from_lambda(self,
@@ -255,7 +257,7 @@ class Environments(collections.abc.Sequence, Sequence[Environment]):
         self._environments = []
 
         for env in environments:
-            if isinstance(env, collections.abc.Sequence):
+            if isinstance(env, (collections.abc.Sequence,collections.abc.Generator)):
                 self._environments.extend(env)
             else:
                 self._environments.append(env)
@@ -265,8 +267,13 @@ class Environments(collections.abc.Sequence, Sequence[Environment]):
         return self.filter(Binary())
 
     def sparse(self, context:bool = True, action:bool = False) -> 'Environments':
-        """Convert an environment from a dense representation to sparse. This has little utility beyond debugging."""
-        return self.filter(Sparse(context,action))
+        """Convert all environments to a sparse representation."""
+        return self.filter(Sparsify(context,action))
+
+    def dense(self, n_feats:int, method:Literal['lookup','hashing'], context:bool = True, action:bool = False) -> 'Environments':
+        """Convert all environments to a dense representation."""
+        make_dense = lambda: Densify(n_feats=n_feats, method=method, context=context, action=action)
+        return Environments([Pipes.join(env, make_dense()) for env in self])
 
     @overload
     def shuffle(self, seed: int = 1) -> 'Environments':
@@ -285,9 +292,10 @@ class Environments(collections.abc.Sequence, Sequence[Environment]):
 
         flat = lambda a: next(pipes.Flatten().filter([a]))
 
-        if kwargs:
+        if kwargs and 'n' in kwargs:
             seeds = range(kwargs['n'])
         else:
+            args = kwargs.get('seed',kwargs.get('seeds',args))        
             seeds = flat(args) or [1]
 
         if isinstance(seeds,int): seeds = [seeds]
@@ -311,7 +319,8 @@ class Environments(collections.abc.Sequence, Sequence[Environment]):
 
     def cycle(self, after: int) -> 'Environments':
         """Cycle all rewards associated with actions by one place."""
-        return self.filter(Cycle(after))
+        if isinstance(after,(int,float)): after = [after]
+        return self.filter([Cycle(a) for a in after])
 
     def params(self, params: Mapping[str,Any]) -> 'Environments':
         """Add params to the environments."""
@@ -320,22 +329,22 @@ class Environments(collections.abc.Sequence, Sequence[Environment]):
     def take(self, n_interactions: int) -> 'Environments':
         """Take a fixed number of interactions from the Environments."""
         return self.filter(Take(n_interactions))
-    
+
     def slice(self, start: Optional[int], stop: Optional[int]=None, step:int = 1) -> 'Environments':
         """Take a slice of interactions from an Environment."""
         return self.filter(Slice(start,stop,step))
 
-    def reservoir(self, n_interactions: int, seeds: Union[int,Sequence[int]]) -> 'Environments':
+    def reservoir(self, n_interactions: int, seeds: Union[int,Sequence[int]]=1) -> 'Environments':
         """Take a random fixed number of interactions from the Environments."""
         if isinstance(seeds,int): seeds = [seeds]
-        return self.filter([Reservoir(n_interactions,seed) for seed in seeds])
+        return self.filter([Reservoir(n_interactions,seed=seed) for seed in seeds])
 
     def scale(self,
         shift: Union[float,Literal["min","mean","med"]] = "min",
         scale: Union[float,Literal["minmax","std","iqr","maxabs"]] = "minmax",
-        targets: Union[Literal["context","rewards","argmax"], Sequence[Literal["context","rewards","argmax"]]] = "context",
+        targets:Literal["context"] = "context",
         using: Optional[int] = None) -> 'Environments':
-        """Apply an affine shift and scaling factor to precondition environments."""
+        """Apply an affine shift and scaling factor to precondition environment features."""
         if isinstance(targets,str): targets = [targets]
         return self.filter(Pipes.join(*[Scale(shift, scale, t, using) for t in targets]))
 
@@ -360,7 +369,7 @@ class Environments(collections.abc.Sequence, Sequence[Environment]):
         reward : Callable[[float,CobaRandom], float] = None,
         seed   : int = 1) -> 'Environments':
         """Add noise to an environment's context, actions and rewards."""
-        return Environments(*self).filter(Noise(context,action,reward,seed))
+        return self.filter(Noise(context,action,reward,seed))
 
     def flat(self) -> 'Environments':
         """Flatten environment's context and actions."""
@@ -368,17 +377,25 @@ class Environments(collections.abc.Sequence, Sequence[Environment]):
 
     def materialize(self) -> 'Environments':
         """Materialize the environments in memory. Ideal for stateful environments such as Jupyter Notebook."""
-        #we use pipes.cache directly because the environment cache will copy 
+        #we use pipes.cache directly because the environment cache will copy
         #which we don't need when we materialize an environment in memory
-        envs = Environments([Pipes.join(env, pipes.Cache(25)) for env in self])
+        envs = Environments([Pipes.join(env, Finalize(apply_repr=False), pipes.Cache(25,True)) for env in self])
+
         for env in envs: list(env.read()) #force read to pre-load cache
+
+        for env in envs:
+            for i in range(len(env)-1):
+                pipe = env[i]
+                if isinstance(pipe, pipes.Cache) and not pipe._protected:
+                    pipe._cache = None
+
         return envs
 
     def grounded(self, n_users: int, n_normal:int, n_words:int, n_good:int, seed:int=1) -> 'Environments':
         """Convert from simulated environments to interaction grounded environments."""
         return self.filter(Grounded(n_users, n_normal, n_words, n_good, seed))
 
-    def repr(self, 
+    def repr(self,
         cat_context:Literal["onehot","onehot_tuple","string"] = "onehot",
         cat_actions:Literal["onehot","onehot_tuple","string"] = "onehot") -> 'Environments':
         """Determine how certain types of data is represented."""
@@ -393,19 +410,23 @@ class Environments(collections.abc.Sequence, Sequence[Environment]):
         envs = Environments([Pipes.join(env, Chunk()) for env in self])
         return envs.cache() if cache else envs
 
-    def logged(self, learners: Union[Learner,Sequence[Learner]], rewards:Literal["DM","IPS"] = "DM", seed:float = 1.23) -> 'Environments':
+    def logged(self, learners: Union[Learner,Sequence[Learner]], seed:Optional[float] = 1.23) -> 'Environments':
         """Create a logged environment using the given learner for the logging policy."""
         if not isinstance(learners, collections.abc.Sequence): learners = [learners]
-        return self.filter(BatchSafe(Finalize())).filter([Logged(learner, rewards, seed) for learner in learners ])
+        return self.filter(BatchSafe(Finalize())).filter([Logged(learner, seed) for learner in learners ])
 
     def unbatch(self):
         """Unbatch interactions in the environments."""
         return self.filter(Unbatch())
 
+    def ope_rewards(self, rewards_type:Literal['IPS','DM','DR'] = None):
+        """Reward estimates for off-policy evaluation."""
+        if isinstance(rewards_type,str): rewards_type = [rewards_type]
+        return self.filter([OpeRewards(r) for r in rewards_type])
+
     def save(self, path: str, processes:int=1, overwrite:bool=False) -> 'Environments':
-
+        """Save the environments to disk."""
         self_envs = list(self)
-
         if Path(path).exists():
             try:
                 path_envs   = Environments.from_save(path)
@@ -429,7 +450,7 @@ class Environments(collections.abc.Sequence, Sequence[Environment]):
                     Path(path).unlink()
                 if not is_equal and not overwrite:
                     raise CobaException("The Environments save file does not match the actual Environments and overwite is False.")
-            
+
             except BadZipFile:
                 if overwrite:
                     Path(path).unlink()
@@ -439,7 +460,7 @@ class Environments(collections.abc.Sequence, Sequence[Environment]):
         CobaContext.logger = DecoratedLogger([ExceptLog()], CobaContext.logger, [StampLog()] if processes ==1 else [NameLog(), StampLog()])
         Pipes.join(IterableSource(self_envs),CobaMultiprocessor(EnvironmentsToObjects(),processes),ObjectsToZipMember(path)).run()
         CobaContext.logger = CobaContext.logger.undecorate()
-        
+
         return Environments.from_save(path)
 
     def cache(self) -> 'Environments':
@@ -451,8 +472,11 @@ class Environments(collections.abc.Sequence, Sequence[Environment]):
         filters = filter if isinstance(filter, collections.abc.Sequence) else [filter]
         return Environments([Pipes.join(e,f) for e in self._environments for f in filters])
 
-    def __getitem__(self, index:int) -> Environment:
-        return self._environments[index]
+    def __getitem__(self, index:Any) -> Union[Environment,'Environments']:
+        if isinstance(index,slice):
+            return Environments(self._environments[index])
+        else:
+            return self._environments[index]
 
     def __iter__(self) -> Iterator[Environment]:
         return iter(self._environments)
