@@ -6,24 +6,25 @@ import copy
 from math import isnan
 from statistics import median, stdev, mode
 from numbers import Number
+from zlib import crc32
 from operator import eq, getitem, methodcaller, itemgetter
-from collections import defaultdict, deque
+from collections import defaultdict
 from functools import lru_cache
 from itertools import islice, chain, tee, compress, repeat
-from typing import Optional, Sequence, Union, Iterable, Any, Tuple, Callable, Mapping
-from coba.backports import Literal
+from typing import Optional, Sequence, Union, Iterable, Any, Tuple, Callable, Mapping, Literal
 
 from coba            import pipes, primitives
 from coba.random     import CobaRandom
 from coba.exceptions import CobaException
 from coba.statistics import iqr
-from coba.utilities  import peek_first
-from coba.primitives import ScaleReward, BinaryReward, SequenceReward, BatchReward, IPSReward
+from coba.utilities  import peek_first, PackageChecker
+from coba.primitives import BinaryReward, SequenceReward, BatchReward, argmax
+from coba.primitives import IPSReward, SequenceFeedback, MappingReward, MulticlassReward
 from coba.primitives import Feedback, BatchFeedback
 from coba.learners   import Learner, SafeLearner
-from coba.pipes      import Filter
+from coba.pipes      import Filter, SparseDense
 
-from coba.environments.primitives import EnvironmentFilter, Interaction
+from coba.environments.primitives import Interaction, EnvironmentFilter, SimpleEnvironment
 
 class Identity(pipes.Identity, EnvironmentFilter):
     """Return whatever interactions are given to the filter."""
@@ -39,12 +40,15 @@ class Slice(pipes.Slice, EnvironmentFilter):
 
 class Shuffle(pipes.Shuffle, EnvironmentFilter):
     """Shuffle a sequence of Interactions in an Environment."""
-    
+
     def filter(self, interactions: Iterable[Interaction]) -> Sequence[Any]:
         first, interactions = peek_first(interactions)
 
-        if 'action' in first and 'reward' in first:
-            #this is here because otherwise offpolicy evaluation can give a
+        if not interactions:
+            yield from []
+
+        elif 'action' in first and 'reward' in first:
+            #this is here because if it is not offpolicy evaluation can give a
             #very biased estimate when seeds are the same. To see this run.
                 # import numpy as np
                 # from coba import CobaRandom
@@ -56,7 +60,14 @@ class Shuffle(pipes.Shuffle, EnvironmentFilter):
                 # R2 = CobaRandom(seed).shuffle(R1)
 
                 # np.corrcoef(R1,R2)
-            yield from CobaRandom("AB").shuffle(list(super().filter(interactions)))
+
+            old_seed = self._seed
+            new_seed = self._seed * 3.21 if self._seed is not None else self._seed
+
+            self._seed = new_seed
+            yield from super().filter(interactions)
+            self._seed = old_seed
+
         else:
             yield from super().filter(interactions)
 
@@ -76,7 +87,7 @@ class Scale(EnvironmentFilter):
     def __init__(self,
         shift: Union[Number,Literal["min","mean","med"]] = 0,
         scale: Union[Number,Literal["minmax","std","iqr","maxabs"]] = "minmax",
-        target: Literal["context","rewards","argmax"] = "context",
+        target: Literal["context"] = "context",
         using: Optional[int] = None):
         """Instantiate a Scale filter.
 
@@ -121,107 +132,72 @@ class Scale(EnvironmentFilter):
         if isinstance(first_context, primitives.Sparse) and self._target=="context" and self._shift != 0:
             raise CobaException("Shift is required to be 0 for sparse environments. Otherwise the environment will become dense.")
 
-        is_discrete       = first_actions and len(first_actions) > 0
         is_dense_context  = isinstance(first_context, primitives.Dense)
         is_sparse_context = isinstance(first_context, primitives.Sparse)
         is_value_context  = not (is_dense_context or is_sparse_context)
 
         #get the values we wish to scale
-        start = time.time()
-        if self._target == "context" and is_dense_context:
-            potential_cols = [i for i,v in enumerate(first['context']) if isinstance(v,(int,float))]
-            if len(potential_cols) == 0:
-                unscaled = []
-            elif len(potential_cols) == 1:
-                unscaled = [ tuple(map(itemgetter(*potential_cols),map(getitem,fitting_interactions,repeat("context")))) ]
-            else:
-                unscaled = list(zip(*map(itemgetter(*potential_cols),map(getitem,fitting_interactions,repeat("context")))))
+        if self._target == "context" :
+            if is_dense_context:
+                potential_cols = [i for i,v in enumerate(first['context']) if isinstance(v,(int,float))]
+                if len(potential_cols) == 0:
+                    unscaled = []
+                elif len(potential_cols) == 1:
+                    unscaled = [ tuple(map(itemgetter(*potential_cols),map(getitem,fitting_interactions,repeat("context")))) ]
+                else:
+                    unscaled = list(zip(*map(itemgetter(*potential_cols),map(getitem,fitting_interactions,repeat("context")))))
+            elif is_sparse_context:
+                unscalable_cols = {k for k,v in first['context'].items() if not isinstance(v,(int,float))}
+                unscaled = defaultdict(list)
+                for interaction in fitting_interactions:
+                    context = interaction['context']
+                    for k in context.keys()-unscalable_cols:
+                        unscaled[k].append(context[k])
+            elif is_value_context:
+                unscaled = [interaction['context'] for interaction in fitting_interactions]
 
-        elif self._target == "context" and is_sparse_context:
-            unscalable_cols = {k for k,v in first['context'].items() if not isinstance(v,(int,float))}
-            unscaled = defaultdict(list)
-            for interaction in fitting_interactions:
-                context = interaction['context']
-                for k in context.keys()-unscalable_cols:
-                    unscaled[k].append(context[k])
-
-        elif self._target == "context" and is_value_context:
-            unscaled = [interaction['context'] for interaction in fitting_interactions]
-
-        elif self._target == "rewards" and is_discrete:
-            rwd_vals = lambda i: list(map(i['rewards'].eval,range(len(i['actions']))))
-            unscaled = sum(list(map(rwd_vals,fitting_interactions)),[])
-
-        elif self._target == "rewards" and not is_discrete:
-            unscaled = []
-
-        elif self._target == "argmax":
-            unscaled = [interaction['rewards'].argmax() for interaction in fitting_interactions]
-        self._times[1] += time.time()-start
-
-        start = time.time()
         #determine the scale and shift values
-        if self._target == "context" and is_dense_context:
-            shifts_scales = []
-            for i,col in zip(potential_cols,unscaled):
-                shift_scale = self._get_shift_and_scale(col)
-                if shift_scale is not None:
-                    shifts_scales.append((i,)+shift_scale)
+        if self._target == "context":
+            if is_dense_context:
+                shifts_scales = []
+                for i,col in zip(potential_cols,unscaled):
+                    shift_scale = self._get_shift_and_scale(col)
+                    if shift_scale is not None:
+                        shifts_scales.append((i,)+shift_scale)
+            elif is_sparse_context:
+                shifts_scales = {}
+                for k,col in unscaled.items():
+                    vals = col + [0]*(len(fitting_interactions)-len(col))
+                    shift_scale = self._get_shift_and_scale(vals)
+                    if shift_scale is not None:
+                        shifts_scales[k] = shift_scale
+            elif is_value_context:
+                shifts_scales = self._get_shift_and_scale(unscaled)
 
-        elif self._target == "context" and is_sparse_context:
-            shifts_scales = {}
-            for k,col in unscaled.items():
-                vals = col + [0]*(len(fitting_interactions)-len(col))
-                shift_scale = self._get_shift_and_scale(vals)
-                if shift_scale is not None:
-                    shifts_scales[k] = shift_scale
-
-        elif self._target in ['context','rewards','argmax']:
-            shifts_scales = self._get_shift_and_scale(unscaled)
-        self._times[2] += time.time()-start
-
-        start = time.time()
+        #now scale
         if not shifts_scales:
             yield from chain(fitting_interactions, remaining_interactions)
-
-        #now scale return
-        elif self._target == "context" and is_dense_context:
-            for interaction in chain(fitting_interactions, remaining_interactions):
-                context = interaction['context']
-                for i,shift,scale in shifts_scales:
-                     context[i] = (context[i]+shift)*scale
-                yield interaction
-
-        elif self._target == "context" and is_sparse_context:
-            for interaction in chain(fitting_interactions, remaining_interactions):
-                new = interaction # Mutable copies
-                context = new['context']
-                for k in shifts_scales.keys() & context.keys():
-                    (shift,scale) = shifts_scales[k]
-                    context[k] = (context[k]+shift)*scale
-                yield new
-
-        elif self._target == "context" and is_value_context:
-            (shift,scale) = shifts_scales
-            for interaction in chain(fitting_interactions, remaining_interactions):
-                new = interaction.copy()
-                new['context'] = (new['context']+shift)*scale
-                yield new
-
-        elif self._target == "rewards":
-            (shift,scale) = shifts_scales
-            for interaction in chain(fitting_interactions, remaining_interactions):
-                new = interaction.copy()
-                new['rewards'] = ScaleReward(new['rewards'], shift, scale, 'value')
-                yield new
-
-        elif self._target == "argmax":
-            (shift,scale) = shifts_scales
-            for interaction in chain(fitting_interactions, remaining_interactions):
-                new = interaction.copy()
-                new['rewards'] = ScaleReward(new['rewards'], shift, scale, 'argmax')
-                yield new
-        self._times[3] += time.time()-start
+        elif self._target == "context":
+            if is_dense_context:
+                for interaction in chain(fitting_interactions, remaining_interactions):
+                    context = interaction['context']
+                    for i,shift,scale in shifts_scales:
+                        context[i] = (context[i]+shift)*scale
+                    yield interaction
+            elif is_sparse_context:
+                for interaction in chain(fitting_interactions, remaining_interactions):
+                    new = interaction # Mutable copies
+                    context = new['context']
+                    for k in shifts_scales.keys() & context.keys():
+                        (shift,scale) = shifts_scales[k]
+                        context[k] = (context[k]+shift)*scale
+                    yield new
+            elif is_value_context:
+                (shift,scale) = shifts_scales
+                for interaction in chain(fitting_interactions, remaining_interactions):
+                    new = interaction.copy()
+                    new['context'] = (new['context']+shift)*scale
+                    yield new
 
     def _get_shift_and_scale(self,values) -> Tuple[float,float]:
         try:
@@ -360,7 +336,7 @@ class Impute(EnvironmentFilter):
                 imputation = self._get_imputation(col)
                 if imputation is not None:
                     imputations[i] = imputation
-                    if self._miss and any([c is None for c in col]): 
+                    if self._miss and any([c is None for c in col]):
                         impute_binary[i] = len(impute_binary)
 
         elif is_sparse:
@@ -388,9 +364,9 @@ class Impute(EnvironmentFilter):
             if is_dense:
                 is_missing = [0]*len(impute_binary)
                 for k,v in enumerate(context):
-                    if v is None and k in imputations: 
+                    if v is None and k in imputations:
                         context[k] = imputations[k]
-                        if k in impute_binary: 
+                        if k in impute_binary:
                             is_missing[impute_binary[k]] = 1
                 context += is_missing
 
@@ -398,7 +374,7 @@ class Impute(EnvironmentFilter):
 
                 is_missing = binary_template.copy()
                 for k,v in context.items():
-                    if v is None: 
+                    if v is None:
                         context[k] = imputations[k]
                         if k in impute_binary:
                             is_missing[impute_binary[k]] = 1
@@ -406,7 +382,7 @@ class Impute(EnvironmentFilter):
 
             elif is_value:
                 if impute_binary:
-                    if context is None: 
+                    if context is None:
                         interaction["context"] = [imputations,1]
                     else:
                         interaction["context"] = [context,0]
@@ -429,14 +405,11 @@ class Impute(EnvironmentFilter):
         except:
             return None
 
-class Sparse(EnvironmentFilter):
-    """Sparsify an environment's feature representation.
-
-    This has little utility beyond debugging.
-    """
+class Sparsify(EnvironmentFilter):
+    """Sparsify an environment's feature representation."""
 
     def __init__(self, context:bool = True, action:bool = False):
-        """Instantiate a Sparse filter.
+        """Instantiate a Sparsify filter.
 
         Args:
             context: If True then contexts should be made sparse otherwise leave them alone.
@@ -448,7 +421,7 @@ class Sparse(EnvironmentFilter):
 
     @property
     def params(self) -> Mapping[str, Any]:
-        return { "sparse_C": self._context, "sparse_A": self._action }
+        return { "sparse_c": self._context, "sparse_a": self._action }
 
     def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
 
@@ -469,20 +442,129 @@ class Sparse(EnvironmentFilter):
                 new['actions'] = list(map(self._make_sparse,new['actions'],repeat(actions_has_headers),repeat('action')))
 
             if self._action and 'action' in new:
-                new['action'] = self._make_sparse(new['action'],action_has_headers, 'action')
+                new['action'] = self._make_sparse(new['action'],action_has_headers,'action')
 
             yield new
 
     def _make_sparse(self, value, has_headers:bool, default_header:str) -> Optional[dict]:
         if value is None:
             return value
-        if isinstance(value,primitives.Dense): 
-            value_list = list(value)
-            return {k:value_list[i] for k,i in value.headers.items() if i < len(value_list)} if has_headers else dict(enumerate(value))
+
+        if isinstance(value,primitives.Dense):
+            if has_headers:
+                value_list = list(value)
+                return {k:value_list[i] for k,i in value.headers.items() if i < len(value_list) and value_list[i] != 0}
+            else:
+                return {str(k):v for k,v in enumerate(value) if v != 0 }
+
         if isinstance(value,primitives.Sparse):
             return value
+
         return {default_header:value}
 
+class Densify(EnvironmentFilter):
+    """Densify an environment's feature representation."""
+
+    def __init__(self,
+        n_feats: int = 400,
+        method : Literal['lookup','hashing'] = 'lookup',
+        context: bool = True,
+        action : bool = False):
+        """Instantiate a Densify filter.
+
+        Args:
+            n_feats: The number of features to use when densifying a sparse entry.
+            method: The method used to convert sparse representation to dense representation.
+            The "lookup" method produces a more efficient feature representation but its memory
+            usage expand to the cardinality of the sparse features. The "hashing" method utilizes
+            a less efficient feature representation but does not use any memory beyond the features.
+            When using the 'hashing' method `n_feats` should be set to a large number such as 2**18.
+            context: If True then contexts should be made sparse otherwise leave them alone.
+            action: If True then actions should be made sparse otherwise leave them alone.
+        """
+
+        self._n_feats = n_feats
+        self._context = context
+        self._action  = action
+        self._method  = method
+
+        def generator():
+            rng = CobaRandom(seed=1)
+            while True:
+                for i in rng.shuffle(range(n_feats)):
+                    yield i
+
+        def factory(g=generator()):
+            return next(g)
+
+        self._lookup  = defaultdict(factory)
+
+    @property
+    def params(self) -> Mapping[str, Any]:
+        return { "dense_m": self._method, "dense_n": self._n_feats, "dense_c": self._context, "dense_a": self._action }
+
+    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
+
+        for interaction in interactions:
+
+            new = interaction.copy()
+
+            if self._context:
+                new['context'] = self._make_dense(new['context'])
+
+            if self._action and 'actions' in new:
+                new['actions'] = list(map(self._make_dense,new['actions']))
+
+            if self._action and 'action' in new:
+                new['action'] = self._make_dense(new['action'])
+
+            yield new
+
+    def _make_dense(self, value) -> Optional[dict]:
+
+        #For this conversion we use the hashing trick. Based on many tests we go with crc32:
+        #   > It is fast
+        #   > It has few collisions
+        #   > It is deterministic between runs
+
+        # Hashing Algorithms We Tested:
+        #   from hashlib import md5, sha1, sha256, sha3_128, sha3_256, shake_128, shake_256, blake2b
+        #   from zlib import adler32, crc32
+        #   from mmh3 import hash as mmh3_hash #(aka, MurmurHash3, requires pip install mmh3)
+        #   from builtins import hash
+
+        # Performance Findings:
+            # hashlib algorithms took between 0.5 to 0.75 seconds to produce 1 million hashes
+            # zlib and mmh3 algorithms took about .16 seconds to produce 1 million hashes
+            # builtin hash algorithm took about .09 seconds to produce 1 million hashes
+
+        # Collision Findings:
+            # we used the code below to count collisions for 1,000,000 randomly generated strings:
+                #random_strings = lambda n: (''.join(map(str,[r*1,r*2,r*3,r*4,r*5])) for r in cb.random.randoms(n))
+                #hashed_values  = [<hash-algo>(s.encode('ascii')) for s in random_strings(1_000_000)]
+                #Counter(Counter(hashed_values).values())
+
+            # hashlib and the builtin hash had 0 collisions
+            # mmh3 and crc32 had ~130 collisions (i.e., 0.013%)
+            # adler32 had ~250k collisions (i.e., 24.00%)
+
+        #Final decisions:
+            # crc32:
+            #   > It is considerably faster than the hashlib algorithms
+            #   > It has equal performance and collisions to MurmurHash3
+            #   > It is deterministic across runs
+
+        if not isinstance(value,primitives.Sparse):
+            return value
+
+        else:
+            if self._method == 'lookup':
+                values = { self._lookup[k]:v for k,v in value.items() }
+            else:
+                values = { crc32(k.encode('ascii')) % self._n_feats: v for k,v in value.items() }
+
+            return SparseDense(values, self._n_feats)
+
 class Cycle(EnvironmentFilter):
     """Cycle all rewards associated with actions by one place.
 
@@ -503,36 +585,37 @@ class Cycle(EnvironmentFilter):
 
     def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
 
-        underlying_iterable     = iter(interactions)
-        sans_cycle_interactions = islice(underlying_iterable, self._after)
-        with_cycle_interactions = underlying_iterable
+        first, interactions = peek_first(interactions)
 
-        yield from sans_cycle_interactions
+        has_actions = first and 'actions' in first
+        has_rewards = first and 'rewards' in first
 
-        first, with_cycle_interactions = peek_first(with_cycle_interactions)
+        one_hot = lambda n,N: tuple([0]*n+[1]+[0]*(N-n-1))
+        rotate  = lambda l: l[-1%n_actions:] + l[:-1%n_actions]
 
-        if with_cycle_interactions:
+        n_actions      = len(first['actions']) if has_actions else 0
+        is_discrete    = 0 < n_actions and n_actions < float('inf')
+        onehot_actions = set(map(one_hot,range(n_actions),repeat(n_actions)))
+        is_onehot      = has_actions and set(first['actions']) == onehot_actions
+        is_categorical = has_actions and all(map(isinstance,first['actions'],repeat(str)))
 
-            action_set = set(first['actions'])
-            n_actions  = len(action_set)
+        is_cyclable = has_actions and has_rewards and is_discrete and (is_onehot or is_categorical)
 
-            onehot_actions  = {tuple([0]*n+[1]+[0]*(n_actions-n-1)) for n in range(n_actions)}
-            is_discrete     = 0 < n_actions and n_actions < float('inf')
-            is_onehots      = action_set == onehot_actions
-            is_categoricals = all([isinstance(a,str) for a in action_set])
+        if not is_cyclable:
+            warnings.warn("Cycle only works for discrete environments without action features. It will be ignored in this case.")
+            yield from interactions
 
-            if not is_discrete or (not is_onehots and not is_categoricals):
-                warnings.warn("Cycle only works for discrete environments without action features. It will be ignored in this case.")
-                yield from with_cycle_interactions
-            else:
-                for interaction in with_cycle_interactions:
-                    rewards = deque(interaction['rewards'])
-                    rewards.rotate(1)
+        else:
+            for i,interaction in enumerate(interactions):
 
-                    new = interaction.copy()
-                    new['rewards'] = SequenceReward(list(rewards))
+                new = interaction.copy()
 
-                    yield new
+                if i >= self._after:
+                    actions = interaction['actions']
+                    rewards = interaction['rewards']
+                    new['rewards'] = SequenceReward(actions,rotate(list(map(rewards.eval,actions))))
+
+                yield new
 
 class Flatten(EnvironmentFilter):
     """Flatten the context and action features for interactions."""
@@ -566,10 +649,18 @@ class Binary(EnvironmentFilter):
         return { "binary": True }
 
     def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
-        for interaction in interactions:
-            new = interaction.copy()
-            new['rewards'] = BinaryReward(interaction['rewards'].argmax())
-            yield new
+
+        first, interactions = peek_first(interactions)
+        is_discrete = 'actions' in first and 0 < len(first['actions']) and len(first['actions']) < float('inf')
+
+        if not is_discrete:
+            warnings.warn("Only discrete action rewards can be made binary. ")
+            yield from interactions
+        else:
+            for interaction in interactions:
+                new = interaction.copy()
+                new['rewards'] = BinaryReward(argmax(interaction['actions'], interaction['rewards']))
+                yield new
 
 class Sort(EnvironmentFilter):
     """Sort a sequence of Interactions in an Environment."""
@@ -734,12 +825,17 @@ class Noise(EnvironmentFilter):
         rng = CobaRandom(self._seed)
 
         for interaction in interactions:
+
             new = interaction.copy()
 
-            noisy_context = self._noises(interaction['context'], rng, self._context_noise)
-            noisy_actions = [self._noises(a, rng, self._action_noise) for a in interaction['actions'] ]
-            noisy_rewards = [ self._noises(r, rng, self._reward_noise) for r in interaction['rewards'] ]
-            noisy_rewards = SequenceReward(noisy_rewards)
+            context = new['context']
+            actions = new['actions']
+            rewards = new['rewards']
+
+            noisy_context = self._noises(context, rng, self._context_noise)
+            noisy_actions = [ self._noises(a, rng, self._action_noise) for a in actions ]
+            noisy_rewards = [ self._noises(r, rng, self._reward_noise) for r in map(rewards.eval,actions) ]
+            noisy_rewards = SequenceReward(noisy_actions, noisy_rewards)
 
             new['context'] = noisy_context
             new['actions'] = noisy_actions
@@ -748,18 +844,14 @@ class Noise(EnvironmentFilter):
             yield new
 
     def _noises(self, value:Union[None,float,str,Mapping,Sequence], rng: CobaRandom, noiser: Callable[[float,CobaRandom], float]):
-
         if isinstance(value, primitives.Sparse):
             #we sort so that noise generation is deterministic with respect to seed
             return { k:self._noise(v, rng, noiser) for k,v in sorted(value.items()) }
-
         if isinstance(value, primitives.Dense):
             return [ self._noise(v, rng, noiser) for v in value ]
-
         return self._noise(value, rng, noiser)
 
     def _noise(self, value:Union[None,float,str], rng: CobaRandom, noiser: Callable[[float,CobaRandom], float]) -> float:
-
         return value if not isinstance(value,(int,float)) else noiser(value, rng)
 
 class Params(EnvironmentFilter):
@@ -788,9 +880,9 @@ class Grounded(EnvironmentFilter):
         def eval(self, arg):
             if not self._rng: self._rng = CobaRandom(self._seed)
             if arg == self._argmax:
-                return self._rng.choice(self._goods) 
+                return self._rng.choice(self._goods)
             else:
-                return self._rng.choice(self._bads) 
+                return self._rng.choice(self._bads)
 
     def __init__(self, n_users: int, n_normal:int, n_words:int, n_good:int, seed:int) -> None:
         self._n_users  = self._cast_int(n_users , "n_users" )
@@ -825,7 +917,7 @@ class Grounded(EnvironmentFilter):
         rng = CobaRandom(self._seed)
 
         #we make it a set for faster contains checks
-        normalids       = set(self.normalids) 
+        normalids       = set(self.normalids)
         normal          = [u in normalids for u in self.userids]
         userid_isnormal = list(zip(self.userids,normal))
 
@@ -833,7 +925,7 @@ class Grounded(EnvironmentFilter):
 
         if not interactions: return []
 
-        is_binary_rwd = {0,1} == set(first['rewards'])
+        is_binary_rwd = set(map(first['rewards'].eval, first['actions'])) == {0,1}
         first_context = first['context']
 
         is_sparse = isinstance(first_context,primitives.Sparse)
@@ -845,17 +937,17 @@ class Grounded(EnvironmentFilter):
         for seed, interaction in enumerate(interactions, self._seed):
 
             userid,normal = rng.choice(userid_isnormal)
-            argmax        = interaction['rewards'].argmax()
+            maxarg        = argmax(interaction['actions'],interaction['rewards'])
 
             new = interaction.copy()
 
             if not is_binary_rwd:
-                new['rewards'] = BinaryReward(argmax)
+                new['rewards'] = BinaryReward(maxarg)
 
             if normal:
-                new['feedbacks'] = Grounded.GroundedFeedback(goods,bads,argmax,seed)
+                new['feedbacks'] = Grounded.GroundedFeedback(goods,bads,maxarg,seed)
             else:
-                new['feedbacks'] = Grounded.GroundedFeedback(bads,goods,argmax,seed)
+                new['feedbacks'] = Grounded.GroundedFeedback(bads,goods,maxarg,seed)
 
             if is_sparse:
                 new['context'] = dict(userid=userid,**new['context'])
@@ -875,15 +967,15 @@ class Grounded(EnvironmentFilter):
         raise CobaException(f"{value_name} must be a whole number and not {value}.")
 
 class Repr(EnvironmentFilter):
-    def __init__(self, 
-        cat_context:Literal["onehot","onehot_tuple","string"] = None,
-        cat_actions:Literal["onehot","onehot_tuple","string"] = None) -> None:
-        self._cat_context = cat_context
-        self._cat_actions = cat_actions
+    def __init__(self,
+        categorical_context:Literal["onehot","onehot_tuple","string"] = None,
+        categorical_actions:Literal["onehot","onehot_tuple","string"] = None) -> None:
+        self._cat_context = categorical_context
+        self._cat_actions = categorical_actions
 
     @property
     def params(self) -> Mapping[str, Any]:
-        return {"cat_context": self._cat_context, "cat_actions": self._cat_actions}
+        return {"categoricals_in_context": self._cat_context, "categoricals_in_actions": self._cat_actions}
 
     def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
 
@@ -891,27 +983,50 @@ class Repr(EnvironmentFilter):
 
         if not interactions: return []
 
-        has_actions = 'actions' in first
+        has_actions   = 'actions' in first and bool(first['actions'])
+        has_rewards   = 'rewards' in first
+        has_feedbacks = 'feedbacks' in first
 
-        I = tee(interactions, 3 if has_actions else 2)
+        n_tee = 1 + int(self._cat_context is not None) + int(self._cat_actions is not None and has_actions)
 
-        interactions     = I[0]
-        cat_context_iter = pipes.EncodeCatRows(self._cat_context).filter(i['context'] for i in I[1])
+        if n_tee == 1:
+            tees = iter([interactions])
+        else:
+            tees = iter(tee(interactions, n_tee))
 
-        if has_actions:
-            cat_actions_iter = pipes.EncodeCatRows(self._cat_actions, True).filter(action for i in I[2] for action in i['actions'])
+        if self._cat_context is not None:
+            cat_context_iter = pipes.EncodeCatRows(self._cat_context).filter(i['context'] for i in next(tees))
+
+        if self._cat_actions and has_actions:
+            cat_actions_iter = pipes.EncodeCatRows(self._cat_actions).filter(i['actions'] for i in next(tees))
+
+        for interaction in next(tees):
 
-        for interaction in interactions:
             new = interaction.copy()
-            new['context'] = next(cat_context_iter)
 
-            if has_actions:
-                new['actions'] = list(islice(cat_actions_iter,len(interaction['actions'])))
+            if self._cat_context:
+                new['context'] = next(cat_context_iter)
+
+            if has_actions and self._cat_actions:
+                new_actions = next(cat_actions_iter)
+                old_actions = new['actions']
+
+                if new_actions != old_actions or type(new_actions[0]) != type(old_actions[0]):
+                    new['actions'] = new_actions
+
+                if new_actions != old_actions:
+                    if has_rewards:
+                        old_rewards = new['rewards']
+                        if isinstance(old_rewards,MulticlassReward):
+                            new['rewards'] = MulticlassReward(new_actions[old_actions.index(argmax(old_actions,old_rewards))])
+                        else:
+                            new['rewards'] = SequenceReward(new_actions,list(map(old_rewards.eval,old_actions)))
+                    if has_feedbacks:
+                        new['feedbacks'] = SequenceFeedback(new_actions,list(map(new['feedbacks'].eval,old_actions)))
 
             yield new
 
 class Batch(EnvironmentFilter):
-
     def __init__(self, batch_size: int) -> None:
         self._batch_size = batch_size
 
@@ -928,8 +1043,9 @@ class Batch(EnvironmentFilter):
             yield new
 
     def _batched(self, iterable, n):
-        "Batch data into lists of length n. The last batch may be shorter."
-        #taken from python itertools recipes
+        #Batch data into lists of length n.
+        #The last batch may be shorter.
+        #Taken from python itertools recipes.
         it = iter(iterable)
         batch = list(islice(it, n))
         while (batch):
@@ -985,6 +1101,9 @@ class BatchSafe(EnvironmentFilter):
 
 class Finalize(EnvironmentFilter):
 
+    def __init__(self, apply_repr: bool = True):
+        self._apply_repr = apply_repr
+
     def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
 
         first, interactions = peek_first(interactions)
@@ -1009,13 +1128,16 @@ class Finalize(EnvironmentFilter):
         actions_materialized = not first_has_actions or (not is_dense_actions and not is_sparse_actions or isinstance(first['actions'][0],(list,tuple,dict)))
         action_materialized  = not first_has_action  or (not is_dense_action  and not is_sparse_action  or isinstance(first['action']    ,(list,tuple,dict)))
 
-        for interaction in Repr("onehot","onehot").filter(interactions):
+        if self._apply_repr:
+            interactions = Repr("onehot","onehot_tuple").filter(interactions)
+
+        for interaction in interactions:
 
             new = interaction.copy()
 
             if not context_materialized and is_dense_context:
                 new['context'] = list(new['context'])
-            elif not context_materialized and is_sparse_context :
+            elif not context_materialized and is_sparse_context:
                 new['context'] = new['context'].copy()
 
             if not actions_materialized and is_dense_actions:
@@ -1037,14 +1159,14 @@ class Chunk(EnvironmentFilter):
 
 class Logged(EnvironmentFilter):
 
-    def __init__(self, learner: Learner, rewards:Literal["DM","IPS"] = "DM", seed:float = 1.23) -> None:
+    def __init__(self, learner: Learner, seed: Optional[float] = 1.23) -> None:
         self._learner = learner
-        self._rewards = rewards
         self._seed    = seed
 
     @property
     def params(self) -> Mapping[str, Any]:
-        return {"learner": SafeLearner(self._learner).params, "logged":True, "rewards": self._rewards, "log_seed":self._seed}
+        learner_params = SafeLearner(self._learner).params
+        return { **learner_params, "learner": learner_params, "logged":True, "log_seed":self._seed}
 
     def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
 
@@ -1056,20 +1178,20 @@ class Logged(EnvironmentFilter):
             raise CobaException("We were unable to create a logged representation of the interaction.")
 
         #Avoid circular dependency
-        from coba.experiments.tasks import SimpleEvaluation
+        from coba.evaluators import OnPolicyEvaluator
+
+        seed = self._seed if self._seed is not None else CobaRandom().random()
 
         I1,I2 = tee(interactions,2)
-        flat_int = Unbatch().filter(I1)
-        eval_log = SimpleEvaluation(record=['action','reward','probability']).process(copy.deepcopy(self._learner),I2,self._seed)
-        for interaction, log in zip(flat_int,eval_log):
+        interactions = Unbatch().filter(I1)
+
+        env = SimpleEnvironment(I2)
+        lrn = copy.deepcopy(self._learner)
+        evaluator = OnPolicyEvaluator(record=['action','reward','probability'],seed=seed)
+
+        for interaction, log in zip(interactions,evaluator.evaluate(env,lrn)):
             out = interaction.copy()
             out.update(log)
-
-            if self._rewards == "IPS":
-                actions = interaction.get('actions',[])
-                action  = actions.index(log['action']) if len(actions) > 0 else log['action']
-                out['rewards'] = IPSReward(log['reward'],action,log['probability'])
-
             yield out
 
 class Mutable(EnvironmentFilter):
@@ -1083,7 +1205,7 @@ class Mutable(EnvironmentFilter):
         first_is_dense   = isinstance(first['context'], primitives.Dense)
         first_is_sparse  = isinstance(first['context'], primitives.Sparse)
         first_is_value   = not (first_is_dense or first_is_sparse)
-        first_is_mutable = isinstance(first['context'], (list,dict))
+        first_is_mutable = isinstance(first['context'], (list,dict,SparseDense))
 
         if first_is_mutable:
             for interaction in interactions:
@@ -1110,3 +1232,75 @@ class Mutable(EnvironmentFilter):
 class MappingToInteraction(Filter[Iterable[Mapping], Iterable[Interaction]]):
     def filter(self, items: Iterable[Mapping]) -> Iterable[Interaction]:
         yield from map(Interaction.from_dict,items)
+
+class OpeRewards(EnvironmentFilter):
+
+    def __init__(self, rwd_type:Literal['IPS','DM','DR']=None):
+        if rwd_type in ['DM','DR']:
+            PackageChecker.vowpalwabbit(f"{rwd_type} OpeRewards")
+        self._rwd_type = rwd_type
+
+    @property
+    def params(self) -> Mapping[str, Any]:
+        return {'ope_reward': self._rwd_type or 'None'}
+
+    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
+        rwd_type = self._rwd_type
+
+        if rwd_type is None:
+            yield from interactions
+
+        elif rwd_type == "IPS":
+            for log in interactions:
+                log = log.copy()
+                log['rewards'] = IPSReward(log['reward'], log['action'], log.get('probability'))
+                yield log
+
+        elif rwd_type in ["DM","DR"]:
+            from coba.learners.vowpal import VowpalMediator
+
+            rng = CobaRandom(1)
+            vw = VowpalMediator()
+
+            # When using cb_adf to estimate rewards for actions instead of the
+            # regression we would receive rewards far outside of the expected boundaries.
+            #vw.init_learner("--cb_adf --interactions xa --interactions xxa --quiet ", label_type=4)
+            vw.init_learner("--interactions xa --interactions xxa --quiet ", label_type=1)
+
+            #so that we don't run forever
+            #when interactions is re-iterable
+            interactions = iter(interactions)
+
+            while True:
+
+                #Batch Shuffling reduces the correlation between
+                #the reward functions learned by this VW learner
+                #and potential downstream VW learners. If this
+                #correlation is strong, we will over estimate
+                #the downstream VW learner's performance.
+                L = list(islice(interactions,4000))
+
+                if not L: break
+
+                for log in rng.shuffle(L):
+                    features = {"x": log['context'], "a": log['action']}
+                    label    = f"{log['reward']} {1/(log['probability'] or 1)}"
+                    vw.learn(vw.make_example(features,label))
+
+                for log in L:
+
+                    examples = vw.make_examples({"x":log['context']}, [{"a":a} for a in log['actions']])
+                    rewards = list(map(vw.predict,examples))
+
+                    if rwd_type=="DR":
+                        log_index = log['actions'].index(log['action'])
+                        rewards[log_index] += (log['reward']-rewards[log_index])/(log['probability'] or 1)
+
+                    log = log.copy()
+
+                    try:
+                        log['rewards'] = MappingReward(dict(zip(log['actions'],rewards)))
+                    except:
+                        log['rewards'] = SequenceReward(log['actions'],rewards)
+
+                    yield log
