@@ -1,13 +1,11 @@
 import pickle
 
 from pathlib import Path
-from collections import abc
 from zipfile import ZipFile, ZIP_DEFLATED
 from itertools import islice, repeat, chain
 from typing import Union, Sequence, Mapping, Iterable, Any
 
-from coba.contexts import CobaContext
-from coba.backports import version
+from coba.context import CobaContext
 from coba.pipes import Source, Sink, Filter
 from coba.environments.primitives import Environment, Interaction
 
@@ -40,18 +38,26 @@ class ZipMemberToObjects(Source[Iterable[object]]):
 
     def read(self) -> Iterable[object]:
         try:
-            with ZipFile(self._zip) as zip:
-                with zip.open(self._member) as f:
-                    yield from map(pickle.load, repeat(f))
+            with ZipFile(self._zip) as z:
+                with z.open(self._member) as m:
+                    yield from map(pickle.load, repeat(m))
         except EOFError:
             pass
 
 class EnvironmentsToObjects(Filter[Environment, Iterable]):
-    def filter(self, envs: Union[Environment,Sequence[Environment]]) -> Iterable[Iterable[object]]:
-        if not isinstance(envs,(abc.Iterable)): envs = [envs]
-        for env in envs:
-            with CobaContext.logger.time("Materializing environment..."):
-                yield list(chain([{"version":1,"coba_version":version("coba")},env.params], env.read()))
+    def filter(self, env: Union[Environment,Sequence[Environment]]) -> Iterable[Iterable[object]]:        
+        with CobaContext.logger.time("Materializing environment..."):
+            yield list(self._env_to_objects(env))
+
+    def _env_to_objects(self,env):
+        from coba import __version__ #imported here to avoid circular dependency
+        yield {"version":2,"coba_version":__version__}
+        yield env.params
+        I = iter(env.read())
+        batch = list(islice(I,1000))
+        while batch: 
+            yield batch
+            batch = list(islice(I,1000))
 
 class EnvironmentFromObjects(Environment):
     def __init__(self, source: Source[Iterable[object]]) -> None:
@@ -59,7 +65,14 @@ class EnvironmentFromObjects(Environment):
 
     @property
     def params(self) -> Mapping[str,Any]:
-        return next(islice(self._source.read(),1,None))
+        return list(islice(self._source.read(),1,2))[0]
 
     def read(self) -> Iterable[Interaction]:
-        yield from islice(self._source.read(),2,None)
+
+        I = iter(self._source.read())
+        version_data = list(islice(I,2))[0]
+        
+        if version_data['version'] == 1: #pragma: no cover
+            yield from I
+        else:
+            yield from chain.from_iterable(I)
