@@ -2,16 +2,15 @@ import math
 
 from statistics import mean
 from itertools import count, islice, cycle
-from typing import Sequence, Dict, Tuple, Any, Callable, Optional, overload, Iterable
-from coba.backports import Literal
+from typing import Sequence, Dict, Tuple, Any, Callable, Optional, overload, Iterable, Literal
 
 from coba.exceptions import CobaException
 from coba.random import CobaRandom
 from coba.encodings import InteractionsEncoder, OneHotEncoder
 
-from coba.environments.primitives import Context, Action, SimulatedEnvironment, SimulatedInteraction, SequenceReward
+from coba.environments.primitives import Context, Action, Environment, SimulatedInteraction, SequenceReward
 
-class LambdaSimulation(SimulatedEnvironment):
+class LambdaSimulation(Environment):
     """A simulation created from generative lambda functions."""
 
     @overload
@@ -58,7 +57,7 @@ class LambdaSimulation(SimulatedEnvironment):
 
     @property
     def params(self) -> Dict[str, Any]:
-        params = { "type": "LambdaSimulation" }
+        params = { "env_type": "LambdaSimulation" }
 
         if hasattr(self, '_seed'):
             params["seed"] = self._seed
@@ -77,7 +76,7 @@ class LambdaSimulation(SimulatedEnvironment):
             actions  = _actions(i, context)
             rewards  = [ _reward(i, context, action) for action in actions]
 
-            yield {'context':context,'actions':actions,'rewards': SequenceReward(rewards) }
+            yield {'context':context,'actions':actions,'rewards': SequenceReward(actions,rewards) }
 
     def __str__(self) -> str:
         return "LambdaSimulation"
@@ -181,7 +180,7 @@ class NeighborsSyntheticSimulation(LambdaSimulation):
 
     @property
     def params(self) -> Dict[str, Any]:
-        return {**super().params, "type": "NeighborsSynthetic", "n_neighborhoods": self._n_neighborhoods }
+        return {**super().params, "env_type": "NeighborsSynthetic", "n_neighborhoods": self._n_neighborhoods }
 
     def __str__(self) -> str:
         return f"NeighborsSynth(A={self._n_actions},c={self._n_context_feats},a={self._n_action_feats},N={self._n_neighborhoods},seed={self._seed})"
@@ -273,7 +272,7 @@ class LinearSyntheticSimulation(LambdaSimulation):
 
     @property
     def params(self) -> Dict[str, Any]:
-        return {**super().params, "type":"LinearSynthetic", "reward_features": self._reward_features }
+        return {**super().params, "env_type":"LinearSynthetic", "reward_features": self._reward_features }
 
     def __reduce__(self) -> Tuple[object, ...]:
         return (LinearSyntheticSimulation, self._args)
@@ -348,7 +347,6 @@ class KernelSyntheticSimulation(LambdaSimulation):
             return [ tuple(rng.randoms(n_action_features,-1.5,1.5)) for _ in range(n_actions) ] if n_action_features else one_hot_acts
 
         def reward(index:int, context:Context, action:Action, rng: CobaRandom) -> float:
-
             part_index = action.index(1) if exemplar_parts > 1                           else 0
             context    = context         if n_context_features                           else []
             action     = action          if n_action_features or not n_context_features  else []
@@ -386,7 +384,7 @@ class KernelSyntheticSimulation(LambdaSimulation):
 
     @property
     def params(self) -> Dict[str, Any]:
-        params = {**super().params, "type": "KernelSynthetic", "n_exemplars": self._n_exemplars, 'kernel': self._kernel}
+        params = {**super().params, "env_type": "KernelSynthetic", "n_exemplars": self._n_exemplars, 'kernel': self._kernel}
 
         if self._kernel == "polynomial":
             params['degree'] = self._degree
@@ -505,7 +503,7 @@ class MLPSyntheticSimulation(LambdaSimulation):
 
     @property
     def params(self) -> Dict[str, Any]:
-        return {**super().params, "type": "MLPSynthetic" }
+        return {**super().params, "env_type": "MLPSynthetic" }
 
     def __reduce__(self) -> Tuple[object, ...]:
         return (MLPSyntheticSimulation, self._args)
