@@ -1,11 +1,11 @@
 import math
 
 from collections import defaultdict
-from typing import Any, Dict, Optional, cast, Hashable, Union
+from typing import Any, Mapping, Sequence, Optional, Hashable
 
 from coba.primitives import Context, Action, Actions
 from coba.statistics import OnlineVariance
-from coba.learners.primitives import Learner, PMF, PMF, PDF, requires_hashables
+from coba.learners.primitives import Learner, PMF, PMF, Prob, requires_hashables
 
 @requires_hashables
 class EpsilonBanditLearner(Learner):
@@ -20,13 +20,23 @@ class EpsilonBanditLearner(Learner):
 
         self._epsilon = epsilon
 
-        self._N: Dict[Hashable, int            ] = defaultdict(int)
-        self._Q: Dict[Hashable, Optional[float]] = defaultdict(int)
+        self._N: Mapping[Hashable, int            ] = defaultdict(int)
+        self._Q: Mapping[Hashable, Optional[float]] = defaultdict(int)
 
     @property
-    def params(self) -> Dict[str, Any]:
+    def params(self) -> Mapping[str, Any]:
         return {"family": "epsilon_bandit", "epsilon": self._epsilon }
 
+    def request(self, context: Context, actions: Actions, request: Actions) -> Sequence[Prob]:
+        values    = [ self._Q[action] for action in actions ]
+        max_value = None if set(values) == {None} else max(v for v in values if v is not None)
+        max_count = sum(v==max_value for v in values)
+
+        prob_selected_randomly = 1/len(actions) * self._epsilon
+        prob_selected_greedily = 1/max_count * (1-self._epsilon)
+
+        return [ prob_selected_randomly + int(self._Q[action]==max_value)*prob_selected_greedily for action in request ]
+
     def predict(self, context: Context, actions: Actions) -> PMF:
         values      = [ self._Q[action] for action in actions ]
         max_value   = None if set(values) == {None} else max(v for v in values if v is not None)
@@ -35,12 +45,12 @@ class EpsilonBanditLearner(Learner):
         prob_selected_randomly = [1/len(actions) * self._epsilon] * len(actions)
         prob_selected_greedily = [ int(i in max_indexes)/len(max_indexes) * (1-self._epsilon) for i in range(len(actions))]
 
-        return PMF([p1+p2 for p1,p2 in zip(prob_selected_randomly,prob_selected_greedily)])
+        return [p1+p2 for p1,p2 in zip(prob_selected_randomly,prob_selected_greedily)]
 
-    def learn(self, context: Context, actions: Actions, action: Action, reward: float, prob: float) -> None:
+    def learn(self, context: Context, action: Action, reward: float, probability: float) -> None:
 
         alpha = 1/(self._N[action]+1)
-        old_Q = cast(float, 0 if self._Q[action] is None else self._Q[action])
+        old_Q = self._Q[action] or 0.0
 
         self._Q[action] = (1-alpha) * old_Q + alpha * reward
         self._N[action] = self._N[action] + 1
@@ -61,19 +71,16 @@ class UcbBanditLearner(Learner):
         """Instantiate a UcbBanditLearner."""
         #these variable names were selected for easier comparison with the original paper
         self._t     : int = 0
-        self._m     : Dict[Action, float         ] = {}
-        self._s     : Dict[Action, int           ] = {}
-        self._v     : Dict[Action, OnlineVariance] = {}
+        self._m     : Mapping[Action, float         ] = {}
+        self._s     : Mapping[Action, int           ] = {}
+        self._v     : Mapping[Action, OnlineVariance] = {}
 
     @property
-    def params(self) -> Dict[str, Any]:
-
+    def params(self) -> Mapping[str, Any]:
         return { "family": "UCB_bandit" }
 
-    def predict(self, context: Context, actions: Actions) -> PMF:
-
-        self._t += 1
-        never_observed_actions = [ a for a in actions if a not in self._m ]
+    def request(self, context: Context, actions: Actions, request: Actions) -> Sequence[Prob]:
+        never_observed_actions = set(actions) - self._m.keys()
 
         if never_observed_actions:
             max_actions = never_observed_actions
@@ -82,9 +89,24 @@ class UcbBanditLearner(Learner):
             max_value   = max(values)
             max_actions = [ a for a,v in zip(actions,values) if v==max_value ]
 
-        return PMF([int(action in max_actions)/len(max_actions) for action in actions])
+        return [int(action in max_actions)/len(max_actions) for action in request]
 
-    def learn(self, context: Context, actions: Actions, action: Action, reward: float, prob: float) -> None:
+    def predict(self, context: Context, actions: Actions) -> PMF:
+
+        never_observed_actions = set(actions) - self._m.keys()
+
+        if never_observed_actions:
+            max_actions = never_observed_actions
+        else:
+            values      = [ self._m[a] + self._Avg_R_UCB(a) for a in actions ]
+            max_value   = max(values)
+            max_actions = [ a for a,v in zip(actions,values) if v==max_value ]
+
+        return [int(action in max_actions)/len(max_actions) for action in actions]
+
+    def learn(self, context: Context, action: Action, reward: float, probability: float) -> None:
+
+        self._t += 1
 
         assert 0 <= reward and reward <= 1, "This algorithm assumes that reward has support in [0,1]."
 
@@ -133,30 +155,29 @@ class UcbBanditLearner(Learner):
 class FixedLearner(Learner):
     """A learner that selects actions according to a fixed distribution."""
 
-    def __init__(self, fixed_pmf_pdf: Union[PMF,PDF]) -> None:
+    def __init__(self, pmf: PMF) -> None:
         """Instantiate a FixedLearner.
 
         Args:
-            fixed_pmf_pdf: A PMF or PDF whose values are the probability of taking each action.
+            pmf: A PMF whose values are the probability of taking each action.
         """
 
-        if callable(fixed_pmf_pdf):
-            self._fixed_pmf = None
-            self._fixed_pdf = fixed_pmf_pdf
-        else:
-            assert round(sum(fixed_pmf_pdf),3) == 1, "The given pmf must sum to one to be a valid pmf."
-            assert all([p >= 0 for p in fixed_pmf_pdf]), "All given probabilities of the pmf must be greater than or equal to 0."
-            self._fixed_pmf = fixed_pmf_pdf
-            self._fixed_pdf = None
+        assert round(sum(pmf),3) == 1, "The given pmf must sum to one to be a valid pmf."
+        assert all([p >= 0 for p in pmf]), "All given probabilities of the pmf must be greater than or equal to 0."
+        self._pmf = pmf
 
     @property
-    def params(self) -> Dict[str, Any]:
+    def params(self) -> Mapping[str, Any]:
         return {"family":"fixed"}
 
-    def predict(self, context: Context, actions: Actions) -> PMF:
-        return PMF(self._fixed_pmf) if self._fixed_pmf else self._fixed_pdf
+    def request(self, context: Context, actions: Actions, request: Actions) -> Sequence[Prob]:
+        probs = self._pmf
+        return [ probs[actions.index(a)] for a in request ]
 
-    def learn(self, context: Context, actions: Actions, action: Action, reward: float, prob: float) -> None:
+    def predict(self, context: Context, actions: Actions) -> PMF:
+        return self._pmf
+
+    def learn(self, context: Context, action: Action, reward: float, prob: float) -> None:
         pass
 
 class RandomLearner(Learner):
@@ -167,11 +188,14 @@ class RandomLearner(Learner):
         pass
 
     @property
-    def params(self) -> Dict[str, Any]:
+    def params(self) -> Mapping[str, Any]:
         return {"family":"random"}
 
-    def predict(self, context: Context, actions: Actions) -> PMF:
-        return PMF([1/len(actions)]*len(actions))
+    def request(self, context: Context, actions: Actions, request: Actions) -> Sequence[Prob]:
+        return [1/len(actions)]*len(request)
 
-    def learn(self, context: Context, actions: Actions, action: Action, reward: float, probability: float) -> None:
+    def predict(self, context: Context, actions: Actions) -> PMF:
+        return [1/len(actions)]*len(actions)
+
+    def learn(self, context: Context, action: Action, reward: float, probability: float) -> None:
         pass
