@@ -1,11 +1,11 @@
-from typing import Any, Dict, Sequence
+from typing import Any, Mapping, Sequence
 
 from coba.exceptions import CobaException
 from coba.utilities import PackageChecker
 from coba.primitives import Context, Action
 from coba.encodings import InteractionsEncoder
 
-from coba.learners.primitives import PMF, Actions, Learner
+from coba.learners.primitives import PMF, Learner, Actions, Prob
 
 class LinUCBLearner(Learner):
     """A contextual bandit learner that represents expected reward as a
@@ -41,7 +41,7 @@ class LinUCBLearner(Learner):
                 are indicated by x's while action features are indicated by a's. For example, xaa means to cross the
                 features between context and actions and actions.
         """
-        PackageChecker.numpy("LinUCBLearner.__init__")
+        PackageChecker.numpy("LinUCBLearner")
 
         self._alpha = alpha
 
@@ -52,12 +52,15 @@ class LinUCBLearner(Learner):
         self._A_inv = None
 
     @property
-    def params(self) -> Dict[str, Any]:
+    def params(self) -> Mapping[str, Any]:
         return {'family': 'LinUCB', 'alpha': self._alpha, 'features': self._X}
 
-    def predict(self, context: Context, actions: Actions) -> PMF:
+    def request(self, context: Context, actions: Actions, request: Actions) -> Sequence[Prob]:
+        probs = self.predict(context,actions)
+        return [ probs[actions.index(a)] for a in request ]
 
-        import numpy as np #type: ignore
+    def predict(self, context: Context, actions: Actions) -> PMF:
+        import numpy as np
 
         if isinstance(actions[0], dict) or isinstance(context, dict):
             raise CobaException("Sparse data cannot be handled by this implementation at this time.")
@@ -66,7 +69,7 @@ class LinUCBLearner(Learner):
             self._X_encoder = InteractionsEncoder(list(set(filter(None,[ f.replace('x','') if isinstance(f,str) else f for f in self._X ]))))
 
         context = context or []
-        features: np.ndarray = np.array([self._X_encoder.encode(x=context,a=action) for action in actions]).T
+        features = np.array([self._X_encoder.encode(x=context,a=action) for action in actions]).T
 
         if(self._A_inv is None):
             self._theta = np.zeros(features.shape[0])
@@ -78,10 +81,9 @@ class LinUCBLearner(Learner):
         action_values = point_estimate + self._alpha*np.sqrt(point_bounds)
         max_indexes   = np.where(action_values == np.amax(action_values))[0]
 
-        return PMF([int(ind in max_indexes)/len(max_indexes) for ind in range(len(actions))])
-
-    def learn(self, context: Context, actions: Actions, action: Action, reward: float, probability: float) -> None:
+        return [int(ind in max_indexes)/len(max_indexes) for ind in range(len(actions))]
 
+    def learn(self, context: Context, action: Action, reward: float, probability: float) -> None:
         import numpy as np
 
         if isinstance(action, dict) or isinstance(context, dict):
@@ -91,7 +93,7 @@ class LinUCBLearner(Learner):
             self._X_encoder = InteractionsEncoder(list(set(filter(None,[ f.replace('x','') if isinstance(f,str) else f for f in self._X ]))))
 
         context = context or []
-        features: np.ndarray = np.array(self._X_encoder.encode(x=context,a=action)).T
+        features = np.array(self._X_encoder.encode(x=context,a=action)).T
 
         if(self._A_inv is None):
             self._theta = np.zeros((features.shape[0]))
