@@ -1,12 +1,11 @@
 import re
 from itertools import repeat, compress
 from sys import platform
-from typing import Any, Dict, Union, Sequence, Mapping, Optional, Tuple
+from typing import Any, Dict, Union, Sequence, Mapping, Optional, Tuple, Literal
 
-from coba.backports import Literal
 from coba.exceptions import CobaException
-from coba.learners.primitives import Learner, PMF
-from coba.primitives import Sparse, Context, Action, Actions
+from coba.learners.primitives import Learner, PMF, Prob
+from coba.primitives import Sparse, Context, Action, Actions, Batch
 from coba.utilities import PackageChecker
 
 Feature       = Union[str,int,float]
@@ -22,10 +21,11 @@ class VowpalMediator:
 
     def __init__(self) -> None:
         self._vw = None
-        self._namespace_keys = []
-        self._namespace_keys_index: Dict[str,int] = {}
- 
-        PackageChecker.vowpalwabbit('VowpalMediator.__init__')
+        self._args = ""
+        self._ns_keys = {}
+        self._ns_keys_cnt = 0
+
+        PackageChecker.vowpalwabbit('VowpalMediator')
 
     @property
     def is_initialized(self) -> bool:
@@ -64,6 +64,7 @@ class VowpalMediator:
         self._vw = pyvw.Workspace(args) if self._version[0] >= 9 else pyvw.vw(args)
         self._label_type = pyvw.LabelType(label_type) if self._version[0] >= 9 else label_type
         self._example_init = pyvw.Example if self._version[0] >= 9 else pyvw.example
+        self._args = args
 
         return self
 
@@ -78,7 +79,7 @@ class VowpalMediator:
         self._vw.learn(example)
         self._vw.finish_example(example)
 
-    def make_example(self, namespaces: Namespaces, label:Optional[str]) -> Any:
+    def make_example(self, namespaces: Namespaces, label:Optional[str] = None) -> Any:
         """Create a VW example.
 
         Args:
@@ -97,18 +98,18 @@ class VowpalMediator:
     def finish(self):
         if self.is_initialized:
             self._vw.finish()
-    
+
     # override to transform example before they get pushed down to VW
     # i.e. change namespaces from the default 'x' in shared to any other namespace
     def transform_example(self, vw_shared, vw_uniques, labels):
         pass
 
-    def make_examples(self, shared: Namespaces, uniques: Sequence[Namespaces], labels:Optional[Sequence[str]]) -> Sequence[Any]:
+    def make_examples(self, shared: Namespaces, uniques: Sequence[Namespaces], labels:Optional[Sequence[str]] = None) -> Sequence[Any]:
         """Create a list of VW examples.
 
         Args:
             shared: The features grouped by namespace in this example.
-            separates: The features, grouped by namespace, unique to each each example. 
+            separates: The features, grouped by namespace, unique to each each example.
             label: An optional label (required if this example will be used for learning).
         """
 
@@ -137,11 +138,11 @@ class VowpalMediator:
                 if not feats and feats != 0 and feats != "":
                     continue
                 elif feats.__class__ is str:
-                    yield (ns, [f"{self._get_namespace_keys(ns,1)[0]}={feats}"])
+                    yield (ns, [f"{self._get_ns_keys(ns,0)}={feats}"])
                 elif feats.__class__ is int or feats.__class__ is float:
-                    yield (ns, [(self._get_namespace_keys(ns,1)[0], feats)])
+                    yield (ns, [(self._get_ns_keys(ns,0), feats)])
                 else:
-                    feats = feats.items() if isinstance(feats, Sparse) else zip(self._get_namespace_keys(ns,len(feats)),feats)
+                    feats = feats.items() if isinstance(feats, Sparse) else zip(self._get_ns_keys(ns,len(feats)),feats)
                     yield (ns, [f"{k}={v}" if v.__class__ is str else (k, v) for k,v in feats if v!= 0])
         else: #pragma: no cover
             #the strange type checks below were faster than traditional methods when performance testing
@@ -149,9 +150,9 @@ class VowpalMediator:
                 if not feats and feats != 0 and feats != "":
                     continue
                 elif feats.__class__ is str:
-                    yield (ns, [f"{self._get_namespace_keys(ns,1)[0]}={feats}"])
+                    yield (ns, {f"{self._get_ns_keys(ns,0)}={feats}":1})
                 elif feats.__class__ is int or feats.__class__ is float:
-                    yield (ns, [(self._get_namespace_keys(ns,1)[0], feats)])
+                    yield (ns, {self._get_ns_keys(ns,0): feats})
                 elif isinstance(feats,Sparse):
                     no_str = not any(map(isinstance,feats.values(),repeat(str))) # most-performant (still 2x slower than no check)
                     #no_str = str not in set(map(attrgetter("__class__"),feats.values()))
@@ -167,32 +168,37 @@ class VowpalMediator:
                                 del new[k]
                         yield (ns,new)
                 else:
-                    d={}
+                    no_str = not any(map(isinstance,compress(feats,feats),repeat(str)))
 
-                    K = self._get_namespace_keys(ns,len(feats))
+                    K = self._get_ns_keys(ns,len(feats))
                     V = feats
 
-                    for k,v in compress(zip(K,V),V):
-                        if v.__class__ is str:
-                            d[f"{k}={v}"] = 1
-                        else:
-                            d[k] = v
+                    if no_str:
+                        d = dict(compress(zip(K,V),V))
+                    else:
+                        d={}
+                        for k,v in compress(zip(K,V),V):
+                            if v.__class__ is str:
+                                d[f"{k}={v}"] = 1
+                            else:
+                                d[k] = v
 
                     yield (ns, d)
 
-    def _get_namespace_keys(self, namespace:str, length:int) -> Sequence[str]:
+    def _get_ns_keys(self, ns:str, length: int) -> Union[str,Sequence[str]]:
 
-        if namespace in self._namespace_keys_index:
-            index = self._namespace_keys_index[namespace]
-            keys  = self._namespace_keys[index:index+length]
-        else:
-            index = len(self._namespace_keys)
-            keys  = list(map(str,range(index,index+length)))
+        if ns not in self._ns_keys:
+            if not length:
+                self._ns_keys[ns] = str(self._ns_keys_cnt)
+                self._ns_keys_cnt += 1
+            else:
+                self._ns_keys[ns] = list(map(str,range(self._ns_keys_cnt,self._ns_keys_cnt+length)))
+                self._ns_keys_cnt += length
 
-            self._namespace_keys_index[namespace] = index
-            self._namespace_keys += keys
+        return self._ns_keys[ns]
 
-        return keys
+    def __str__(self) -> str:
+        return self._args
 
 class VowpalLearner(Learner):
     """A friendly wrapper around Vowpal Wabbit's python interface to support CB learning.
@@ -268,7 +274,7 @@ class VowpalLearner(Learner):
         if not self._adf:
             n_action_str = re.match("--cb.*?\s*(\d*)\\b.*", args).group(1)
             self._n_actions = int(n_action_str) if n_action_str else None
-            #useful for removing the n_actions from args 
+            #useful for removing the n_actions from args
             #re.sub("(--cb)\s*(\d+)", '\g<1>', "--cb 123", count=1)
 
         self._vw = vw or VowpalMediator()
@@ -277,8 +283,15 @@ class VowpalLearner(Learner):
     def params(self) -> Mapping[str, Any]:
         return {"family": "vw", 'args': self._args.replace("--quiet","").strip()}
 
+    def request(self, context: Context, actions: Actions, request: Actions) -> Sequence[Prob]:
+        probs = self.predict(context,actions)[0]
+        return probs if actions == request else list(map(probs.__getitem__,map(actions.index,request)))
+
     def predict(self, context: Context, actions: Sequence[Action]) -> PMF:
 
+        if not self._vw.is_initialized and isinstance(context,Batch):#pragma: no cover
+            raise CobaException("VW learner does not support batched calls.")
+
         if not self._vw.is_initialized and self._adf:
             self._vw.init_learner(self._args, 4)
 
@@ -323,9 +336,9 @@ class VowpalLearner(Learner):
             index = self._vw.predict(self._vw.make_example(context, None))
             probs = [ int(i==index) for i in range(1,len(actions)+1) ]
 
-        return PMF(probs)
+        return probs, {'actions':actions}
 
-    def learn(self, context: Context, actions:Actions, action: Action, reward: float, probability: float) -> None:
+    def learn(self, context: Context, action: Action, reward: float, probability: float, actions: Actions = None) -> None:
 
         if not self._vw.is_initialized and self._adf:
             self._vw.init_learner(self._args, 4)
@@ -336,6 +349,8 @@ class VowpalLearner(Learner):
         if not self._vw.is_initialized and not self._adf and not self._n_actions:
             raise CobaException("When using `cb` without `adf` predict must be called before learn to initialize the vw learner")
 
+        if actions is None: actions = [action]
+
         index   = actions.index(action)
         labels  = self._labels(actions, index, reward, probability)
         label   = labels[index]
@@ -349,7 +364,9 @@ class VowpalLearner(Learner):
             self._vw.learn(self._vw.make_example(context, label))
 
     def _labels(self,actions,index,reward:float,prob:float) -> Sequence[Optional[str]]:
-        return [ f"{i+1}:{round(-reward,5)}:{round(prob,5)}" if i == index else None for i in range(len(actions))]
+        labels = [None]*len(actions)
+        labels[index] = f"{index+1}:{round(-reward,5)}:{round(prob,5)}"
+        return labels
 
     def _finish(self):
         try:
@@ -505,7 +522,6 @@ class VowpalCoverLearner(VowpalLearner):
         vw_args_string = " ".join(self.make_args(namespace_interactions=features, vw_kwargs=vw_kwargs))
         super().__init__(vw_args_string, vw)
 
-
 class VowpalRndLearner(VowpalLearner):
     def __init__(self,
                  rnd: int = 3,
@@ -545,7 +561,6 @@ class VowpalRndLearner(VowpalLearner):
         vw_args_string = " ".join(self.make_args(namespace_interactions=features, vw_kwargs=vw_kwargs))
         super().__init__(vw_args_string, vw)
 
-
 class VowpalRegcbLearner(VowpalLearner):
     """A wrapper around VowpalLearner that provides more documentation. For more
         information on the types of exploration algorithms availabe in VW see `here`__.
@@ -635,9 +650,9 @@ class VowpalOffPolicyLearner(VowpalLearner):
     """A wrapper around VowpalLearner that provides more documentation. For more
         information on the types of exploration algorithms availabe in VW see `here`__.
 
-        This wrapper in particular performs policy learning without any exploration. This is
-        only correct when training examples come from a logging policy so that any exploration
-        on our part would be irrelevant.
+        This wrapper performs policy learning without any exploration. This is only correct
+        when training examples come from a logging policy so that any exploration on our
+        part would be irrelevant.
 
         __ https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Contextual-Bandit-algorithms
     """
