@@ -1,11 +1,12 @@
 import pickle
+import importlib.util
 import threading as mt
 import multiprocessing as mp
 
+from collections.abc import Iterator
 from queue import Empty
 from traceback import format_tb
-from typing import Iterable, Mapping, Callable, Optional, Union, Sequence, Any
-from coba.backports import Literal
+from typing import Iterable, Mapping, Callable, Optional, Union, Sequence, Any, Literal
 
 from coba.utilities import peek_first
 from coba.exceptions import CobaException
@@ -29,10 +30,10 @@ from coba.pipes.sinks import QueueSink
 # handle Experiment.evaluate not being called inside of __name__=='__main__' (manually tested)
 #   > Handled by checking exitcode on processes and using an event synchronizer on first process start
 
-#There are three potential contexts -- spawn, fork, and forkserver. On windows and mac spawn is the only option.
-#On Linux the default option is fork. Fork creates processes faster and can share memory but also doesn't play
-#well with threads. Therefore, to make behavior consistent across Linux, Windows, and Mac and avoid potential bugs
-#we force our multiprocessors to always use spawn.
+# There are three potential contexts -- spawn, fork, and forkserver. On Windows and Mac spawn is the only option.
+# On Linux the default option is fork. Fork creates processes faster and can share memory but also doesn't play
+# well with threads. Therefore, to make behavior consistent across Linux, Windows, and Mac -- and to avoid
+# potential bugs -- we force our multiprocessors to always use spawn regardless of the host OS.
 spawn_context = mp.get_context("spawn")
 
 class ProcessLine(spawn_context.Process):
@@ -54,6 +55,7 @@ class ProcessLine(spawn_context.Process):
         self._callback         = None
         self._recv, self._send = spawn_context.Pipe(False)
         self._lock             = spawn_context.Lock()
+
         super().start()
 
         if callback:
@@ -71,10 +73,11 @@ class ProcessLine(spawn_context.Process):
                     "We attempted to evaluate your code in multiple processes but we were unable to find all the code "
                     "definitions needed to pass the tasks to the processes. The two most common causes of this error are: "
                     "1) a learner or simulation is defined in a Jupyter Notebook cell or 2) a necessary class definition "
-                    "exists inside the `__name__=='__main__'` code block in the main execution script. In either case "
-                    "you can choose one of three simple solutions: 1) evaluate your code on a single process with no limit on "
-                    "child tasks, 2) if in Jupyter notebook define all necessary classes in a separate file and include the "
-                    "classes via import statements, or 3) move your class definitions outside the `__name__=='__main__'` check."
+                    "exists inside the `__name__=='__main__'` code block in the main execution script. In either case you "
+                    "can choose one of four simple solutions: 1) pip install cloudpickle into your python environment, 2) "
+                    "evaluate your code on a single process, 3) if in Jupyter notebook define all necessary classes in a "
+                    "separate file and include the classes via import statements, or 4) move your class definitions outside "
+                    "the `__name__ == '__main__'` check."
                 ),None
             else:
                 ex,tb = e,format_tb(e.__traceback__)
@@ -164,7 +167,7 @@ class ThreadLine(mt.Thread):
 
 class AsyncableLine(SourceSink):
 
-    def run_async(self, 
+    def run_async(self,
         callback:Callable[['AsyncableLine',Optional[Exception],Optional[str]],None]=None,
         mode:Literal['process','thread']='process') -> Union[ThreadLine,ProcessLine]:
         """Run the pipeline asynchronously."""
@@ -181,27 +184,54 @@ class AsyncableLine(SourceSink):
         worker.start()
         return worker
 
-class Unchunker:
-    def __init__(self, chunked: bool) -> None:
-        self._chunked = chunked
+class Safe:
+    def __init__(self, filter: Filter):
+        if importlib.util.find_spec('cloudpickle'):
+            import cloudpickle
+            try:
+                self._filter = cloudpickle.dumps(filter)
+            except:
+                self._filter = filter
+        else:
+            self._filter = filter
+
+    def filter(self,items):
+        if importlib.util.find_spec('cloudpickle') and isinstance(self._filter,bytes):
+            import cloudpickle
+            filter = cloudpickle.loads(self._filter)
+        else:
+            filter = self._filter
+
+        yield from filter.filter(items)
+
+class Foreach:
+    def __init__(self,pipe:Filter) -> None:
+        self._pipe = pipe
+
     def filter(self, items: Iterable[Any]) -> Iterable[Any]:
-        if not self._chunked: items = [items]
-        for item in items: yield from item
+        for item in items:
+            out = self._pipe.filter(item)
+            out = out if isinstance(out,Iterator) else [out]
+            yield from out
 
 class Pickler:
-
     def filter(self, items) -> Iterable[bytes]:
         try:
-            yield from map(pickle.dumps,items)
+            if importlib.util.find_spec('cloudpickle'):
+                import cloudpickle
+                yield from map(cloudpickle.dumps,items)
+            else:
+                yield from map(pickle.dumps,items)
         except Exception as e:
             if "pickle" in str(e) or "Pickling" in str(e):
                 message = (
                     f"We attempted to process your code on multiple processes but were unable to do so due to a pickle "
-                    f"error. The exact error received was '{str(e)}'. Errors this kind can often be fixed in one of two "
-                    f"ways: 1) evaluate the experiment in question on a single process with no limit on the tasks per child "
-                    f"or 2) modify the named class to be picklable. The easiest way to make a given class picklable is to "
-                    f"add `def __reduce__(self): return (<the class in question>, (<tuple of constructor arguments>))` to "
-                    f"the class. For more information see https://docs.python.org/3/library/pickle.html#object.__reduce__."
+                    f"error. The exact error received was '{str(e)}'. Errors of this kind can often be fixed in one of three "
+                    f"ways: 1) pip install cloudpickle in your conda environment, 2) evaluate the experiment in question on "
+                    f"a single process with no limit on the tasks per child or 3) modify the named class to be picklable. "
+                    f"The easiest way to make a given class picklable is to add `def __reduce__(self): return (<the class "
+                    f"in question>, (<tuple of constructor arguments>))` to the class. For more information see "
+                    f"https://docs.python.org/3/library/pickle.html#object.__reduce__."
                 )
                 raise CobaException(message)
             else: #pragma: no cover
@@ -209,7 +239,11 @@ class Pickler:
 
 class Unpickler:
     def filter(self, items):
-        yield from map(pickle.loads,items)
+        if importlib.util.find_spec('cloudpickle'):
+            import cloudpickle
+            yield from map(cloudpickle.loads,items)
+        else:
+            yield from map(pickle.loads,items)
 
 class EventSetter:
     def __init__(self, event: mt.Event) -> None:
@@ -236,18 +270,15 @@ class Multiprocessor(Filter[Iterable[Any], Iterable[Any]]):
     def __init__(self,
             filter: Filter[Iterable[Any], Iterable[Any]],
             n_processes: int = 1,
-            maxtasksperchild: int = 0,
-            chunked: bool = False,) -> None:
+            maxtasksperchild: int = 0) -> None:
         """Instantiate a Multiprocessor.
 
         Args:
             filter: The inner pipe that will be executed on multiple processes.
             n_processes: The number of processes that should be created to filter items.
             maxtasksperchild: The number of items/chunks a process should filter before restarting.
-            chunked: Indicates that the given items have been chunked.
         """
         self._filter           = filter
-        self._chunked          = chunked
         self._max_processes    = n_processes
         self._maxtasksperchild = maxtasksperchild or None
 
@@ -261,10 +292,10 @@ class Multiprocessor(Filter[Iterable[Any], Iterable[Any]]):
         if not items: return []
 
         if self._max_processes == 1 and self._maxtasksperchild is None:
-            yield from self._filter.filter(Unchunker(self._chunked).filter(items))
-        else:
+            yield from Foreach(self._filter).filter(items)
 
-            event     = spawn_context.Event()
+        else:
+            event = spawn_context.Event()
 
             #for some reason if this mp queue get too big we can't keyboradinterrupt
             #therefore, we slightly limit its size and then empty it before closing.
@@ -275,9 +306,8 @@ class Multiprocessor(Filter[Iterable[Any], Iterable[Any]]):
             out_put   = QueueSink(out_queue,foreach=True)
             out_get   = QueueSource(out_queue)
             pickler   = Pickler()
-            unpickler = Unpickler() 
+            unpickler = Unpickler()
             get_max   = Slice(None,self._maxtasksperchild)
-            unchunk   = Unchunker(self._chunked)
             setter    = EventSetter(event)
 
             self._n_procs      = self._max_processes
@@ -287,7 +317,7 @@ class Multiprocessor(Filter[Iterable[Any], Iterable[Any]]):
             self._load_stopper = Stopper() #this works because the loader is a thread which means we have shared memory
 
             load_line   = SourceSink(IterableSource(items), self._load_stopper, pickler, in_put)
-            filter_line = SourceSink(in_get, setter, unpickler, get_max, unchunk, self._filter, out_put)
+            filter_line = SourceSink(in_get, setter, unpickler, get_max, Safe(Foreach(self._filter)), out_put)
 
             def loader_finished_or_failed(worker: Union[ThreadLine,ProcessLine]):
                 if worker.exception: self._exceptions.append(worker.exception)
@@ -300,10 +330,13 @@ class Multiprocessor(Filter[Iterable[Any], Iterable[Any]]):
 
                 #only known cause of exitcode != 0 is a missing `if __name__ == '__main__'``.
                 if worker.exitcode != 0: #pragma: no cover
+                    #exitcode -15 is keyboard interrupt...
+                    if worker.exitcode != -15:
+                        print(worker.exitcode)
                     self._main_err = True
                     event.set()
 
-                #we have to stop on exception because depending on where the exception occurred
+                #we have to stop on exception since, depending on where the exception occurred,
                 #we may not have actually read anything from the input queue. If we didn't then
                 #the input queue will never empty and we'll be stuck starting processes forever.
                 if not worker.poisoned and not self._exceptions and worker.exitcode == 0:
@@ -342,7 +375,7 @@ class Multiprocessor(Filter[Iterable[Any], Iterable[Any]]):
                     while True: in_queue.get_nowait()
                 except Empty:
                     pass
-                    #closing can cause exceptions 
+                    #closing can cause exceptions
                     #and doesn't seem to help anything
                     #in_queue.close()
 
