@@ -1,7 +1,6 @@
 from collections import abc
 from itertools import count, compress, chain, filterfalse, islice, repeat
-from typing import Any, Union, Callable, Iterator, Sequence, Mapping, Iterable, Tuple
-from coba.backports import Literal
+from typing import Dict, Any, Union, Callable, Iterator, Sequence, Mapping, Iterable, Tuple, Literal
 
 from coba.primitives import Sparse, Dense, Categorical
 from coba.primitives.rows import Dense_, Sparse_
@@ -33,7 +32,6 @@ class LazyDense(Dense_):
 
         if not enc:
             return val
-
         else:
             try:
                 return enc[key](val)
@@ -59,7 +57,7 @@ class LazyDense(Dense_):
             try:
                 yield e(v)
             except:
-                if v in ['?','']: 
+                if v in ['?','']:
                     yield None
                 else:
                     raise
@@ -85,14 +83,14 @@ class LazySparse(Sparse_):
     def __getitem__(self, key: str):
         key = self._fwd.get(key,key)
         enc = self._enc
-        
+
         try:
             val = self._load_or_get()[key]
         except KeyError:
             if key not in self._nsp: raise
             val = "0"
 
-        if not enc: 
+        if not enc:
             return val
         else:
             try:
@@ -122,7 +120,7 @@ class LazySparse(Sparse_):
             return tuple((inv.get(k,k), v) for k,v in row.items())
         else:
             return tuple(row.items())
-        
+
     def _enc_items(self)->Iterator:
         enc   = self._enc
         inv   = self._inv
@@ -138,6 +136,47 @@ class LazySparse(Sparse_):
 
             yield (inv.get(k,k) if inv else k,v)
 
+class SparseDense(Dense_):
+    __slots__=('_values','_length')
+
+    def __init__(self, values: Dict[int,Any], length:int) -> None:
+        self._values = values
+        self._length = length
+
+    def copy(self) -> 'SparseDense':
+        return SparseDense(self._values.copy(), self._length)
+
+    def __setitem__(self, key:int, value:Any):
+        key = key if key >= 0 else key+self._length
+
+        if key < 0 or key >= self._length:
+            raise IndexError("list index out of range")
+
+        self._values[key] = value
+
+    def __getitem__(self, key: int):
+        key = key if key >= 0 else key+self._length
+
+        if key < 0 or key >= self._length:
+            raise IndexError("list index out of range")
+
+        return self._values.get(key,0)
+
+    def __iter__(self) -> Iterator:
+        sort = sorted(self._values.items())
+
+        yield from repeat(0,sort[0][0])
+        yield sort[0][1]
+
+        for p,n in zip(sort,sort[1:]):
+            yield from repeat(0,n[0]-p[0]-1)
+            yield n[1]
+
+        yield from repeat(0,self._length-sort[-1][0]-1)
+
+    def __len__(self) -> int:
+        return self._length
+
 class HeadDense(Dense_):
     __slots__=('_row','headers')
 
@@ -173,11 +212,11 @@ class HeadSparse(Sparse_):
         return len(self._row)
 
     def keys(self) -> abc.KeysView:
-        head_map_inv_get = self._inv.__getitem__ 
+        head_map_inv_get = self._inv.__getitem__
         return set(map(head_map_inv_get, self._row.keys()))
 
     def items(self) -> Sequence:
-        head_map_inv_get = self._inv.__getitem__ 
+        head_map_inv_get = self._inv.__getitem__
         return tuple((head_map_inv_get(k),v) for k,v in self._row.items())
 
 class HeadRows(Filter[Iterable[Union[Dense,Sparse]],Iterable[Union[Dense,Sparse]]]):
@@ -310,7 +349,7 @@ class KeepDense(Dense_):
 
 class DropSparse(Sparse_):
     __slots__=('_row','_drop_set')
-    
+
     def __init__(self, row: Sparse, drop_set: set = None) -> None:
         self._row      = row
         self._drop_set = drop_set
@@ -334,7 +373,7 @@ class DropSparse(Sparse_):
 
     def items(self) -> Sequence:
         drop = self._drop_set
-        yield from (item for item in self._row.items() if item[0] not in drop) 
+        yield from (item for item in self._row.items() if item[0] not in drop)
 
 class DropRows(Filter[Iterable[Union[Dense,Sparse]], Iterable[Union[Dense,Sparse]]]):
     """A filter which drops rows and columns from in table shaped data."""
@@ -367,7 +406,7 @@ class DropRows(Filter[Iterable[Union[Dense,Sparse]], Iterable[Union[Dense,Sparse
             mapping = {k:v for k,v in chain(enumerate(indexes),headers)}
             length  = len(indexes)
 
-            if headers: 
+            if headers:
                 external_indexes = dict(zip(indexes,count()))
                 external_headers = { h: external_indexes[i] for h,i in headers if i in external_indexes }
             else:
@@ -493,64 +532,60 @@ class LabelRows(Filter[Iterable[Union[Dense,Sparse]],Iterable[Union[Dense,Sparse
             return map(LabelSparse, rows, repeat(label), repeat(tipe))
 
 class EncodeCatRows(Filter[Iterable[Union[Any,Dense,Sparse]], Iterable[Union[Any,Dense,Sparse]]]):
-    def __init__(self, tipe=Literal["onehot","onehot_tuple","string"], value_rows:bool = False) -> None:
+    def __init__(self, tipe=Literal["onehot","onehot_unflat","string"], value_rows:bool = False) -> None:
         self._tipe = tipe
-        self._value_rows = value_rows
 
     def filter(self, rows: Iterable[Union[Any,Dense,Sparse]]) -> Iterable[Union[Any,Dense,Sparse]]:
-
         if self._tipe is None: return rows
         first, rows = peek_first(rows)
         if not rows: return []
-    
-        if self._value_rows:
-            rows = self._encode_value_generator(rows)
-        elif isinstance(first,Dense):
-            rows = self._encode_dense_generator(rows, first)
-        elif isinstance(first,Sparse):
-            rows = self._encode_sparse_generator(rows, first)
+
+        if isinstance(first,(Dense,Sparse)):
+            rows = self._encode_collection(rows, first)
         else:
-            rows = rows
+            rows = self._encode_values(rows, first)
 
         if self._tipe =='onehot':
             rows = Flatten().filter(rows)
 
         return rows
 
-    def _encode_value_generator(self, rows):
-        first,rows = peek_first(rows)
+    def _encode_values(self, rows, first):
         if not isinstance(first,Categorical):
             yield from rows
-        elif self._tipe == "string":
+        elif self._tipe == 'string':
             yield from map(str,rows)
-        elif "onehot" in self._tipe:
-            for row in rows: yield row.as_onehot
+        else:
+            yield from (r.as_onehot for r in rows)
 
-    def _encode_dense_generator(self, rows, first):
-        is_mutable = isinstance(first, list)
-        cat_cols =  [i for i,v in enumerate(first) if isinstance(v,Categorical) ]
-       
-        if self._tipe == "string":
-            for row in rows:
-                row = row.copy() if is_mutable else list(row)
-                for i in cat_cols: row[i] = str(row[i])
-                yield row
-        elif "onehot" in self._tipe:
-            for row in rows:
-                row = row.copy() if is_mutable else list(row)
-                for i in cat_cols: row[i] = row[i].as_onehot
-                yield row
+    def _encode_collection(self, rows, first):
+
+        if isinstance(first,Sparse):
+            cat_cols = [k for k,v in first.items() if isinstance(v,Categorical) ]
+        else:
+            cat_cols = [i for i,v in enumerate(first) if isinstance(v,Categorical) ]
+
+        copyable = not isinstance(first,tuple)
+        get_string = 'string' == self._tipe
+
+        prev_row   = None
+        prev_yield = None
+
+        for row in rows:
+            if row == prev_row:
+                #this check doesn't cost us much and will greatly
+                #speed up the encoding of repeated categorical actions
+                yield prev_yield
+            else:
+                prev_row = row
+                row = row.copy() if copyable else list(row)
+
+                if get_string:
+                    for k in cat_cols:
+                        row[k] = str(row[k])
+                else:
+                    for k in cat_cols:
+                        row[k] = row[k].as_onehot
 
-    def _encode_sparse_generator(self, rows, first):
-        cat_cols = [k for k,v in first.items() if isinstance(v,Categorical) ]
-        
-        if self._tipe == "string":
-            for row in rows:
-                row = row.copy()
-                for k in cat_cols: row[k] = str(row[k])
-                yield row
-        elif "onehot" in self._tipe:
-            for row in rows:
-                row = row.copy()
-                for k in cat_cols: row[k] = row[k].as_onehot
                 yield row
+                prev_yield = row
