@@ -83,11 +83,12 @@ class CobaRandom:
         """
         return min+(max-min)*next(self._randu)        
 
-    def shuffle(self, sequence: Sequence[Any]) -> Sequence[Any]:
+    def shuffle(self, sequence: Sequence[Any], inplace: bool = False) -> Sequence[Any]:
         """Shuffle the order of items in a sequence.
 
         Args:
             sequence: The sequence of items that are to be shuffled.
+            inplace: The sequence of items should be shuffled within the given sequence
 
         Returns:
             A new sequence with the order of items shuffled.
@@ -99,13 +100,15 @@ class CobaRandom:
 
         n = len(sequence)
         if n < 2: return sequence
-        l = list(sequence)
+
+        l = sequence if inplace else list(sequence)
 
         #i goes from 0 to n-2
         #j is always i <= j < n
         for i,r in islice(enumerate(self._randu),n-1):
             j = i+int(r*(n-i)) 
             l[i], l[j] = l[j], l[i]
+        
         return l
 
     def randint(self, a:int, b:int) -> int:
@@ -306,14 +309,15 @@ def choice(seq: Sequence[Any], weights:Sequence[float]=None) -> Any:
 
     return _random.choice(seq, weights)
 
-def shuffle(array_like: Sequence[Any]) -> Sequence[Any]:
+def shuffle(array_like: Sequence[Any], inplace:bool=False) -> Sequence[Any]:
     """Shuffle the order of items in a sequence.
 
     Args:
         sequence: The sequence of items that are to be shuffled.
+        inplace: The sequence of items should be shuffled within the given sequence.
 
     Returns:
         A new sequence with the order of items shuffled.
     """
 
-    return _random.shuffle(array_like)
+    return _random.shuffle(array_like, inplace)
