@@ -1,8 +1,8 @@
 from coba.registry     import CobaRegistry
 from coba.pipes        import NullSink, ConsoleSink, DiskSink, HttpSource
 from coba.environments import OpenmlSimulation, SupervisedSimulation
-from coba.environments import Sort, Scale, Cycle, Shuffle, Take, Identity, Where, Repr
-from coba.contexts     import DiskCacher, NullCacher, IndentLogger, NullLogger, BasicLogger
+from coba.environments import Sort, Scale, Cycle, Shuffle, Take, Identity, Where, Repr, Reservoir
+from coba.context      import DiskCacher, NullCacher, IndentLogger, NullLogger, BasicLogger
 
 CobaRegistry.register("range" , range)
 CobaRegistry.register("zip"   , zip  )
@@ -22,11 +22,12 @@ CobaRegistry.register("NullCacher", NullCacher)
 CobaRegistry.register("OpenmlSimulation"    , OpenmlSimulation    )
 CobaRegistry.register("SupervisedSimulation", SupervisedSimulation)
 
-CobaRegistry.register("Identity", Identity)
-CobaRegistry.register("Take"    , Take    )
-CobaRegistry.register("Shuffle" , Shuffle )
-CobaRegistry.register("Sort"    , Sort    )
-CobaRegistry.register("Scale"   , Scale   )
-CobaRegistry.register("Cycle"   , Cycle   )
-CobaRegistry.register("Where"   , Where   )
-CobaRegistry.register("Repr"    , Repr    )
+CobaRegistry.register("Identity" , Identity )
+CobaRegistry.register("Take"     , Take     )
+CobaRegistry.register("Reservoir", Reservoir)
+CobaRegistry.register("Shuffle"  , Shuffle  )
+CobaRegistry.register("Sort"     , Sort     )
+CobaRegistry.register("Scale"    , Scale    )
+CobaRegistry.register("Cycle"    , Cycle    )
+CobaRegistry.register("Where"    , Where    )
+CobaRegistry.register("Repr"     , Repr     )
