@@ -1,13 +1,13 @@
 from math import hypot, isnan, erf, sqrt
+from statistics import fmean
 from sys import version_info
 from operator import mul, sub
-from itertools import repeat
+from bisect import bisect_left
+from itertools import repeat, accumulate, compress, chain
 from abc import abstractmethod, ABC
-from typing import Sequence, Tuple, Union, Callable
-from coba.backports import Literal
+from typing import Sequence, Tuple, Union, Callable, Optional, Literal
 
 from coba.exceptions import CobaException
-from coba.random import CobaRandom
 from coba.utilities import PackageChecker
 
 def iqr(values: Sequence[float]) -> float:
@@ -20,32 +20,63 @@ def iqr(values: Sequence[float]) -> float:
 
     return p75-p25
 
-def percentile(values: Sequence[float], percentiles: Union[float,Sequence[float]]) -> Union[float, Tuple[float,...]]:
+def percentile(values: Sequence[float], percentiles: Union[float,Sequence[float]], weights: Sequence[float] = None, sort: bool = True) -> Union[float, Tuple[float,...]]:
+
+    if len(values) == 1:
+        if isinstance(percentiles,(int,float)):
+            return values[0]
+        else:
+            return list(values)*len(percentiles)
+
+    def _percentile(values: Sequence[float], weights:Optional[Sequence[float]], percentile: float) -> float:
 
-    def _percentile(values: Sequence[float], percentile: float) -> float:
         assert 0 <= percentile and percentile <= 1, "Percentile must be between 0 and 1 inclusive."
 
-        i = percentile*(len(values)-1)
+        if percentile == 0:
+            return values[0]
 
-        if i == int(i):
-            return values[int(i)]
+        if percentile == 1:
+            return values[-1]
+
+        if weights:
+            R = bisect_left(weights,percentile)
+            L = R-1
+            LP = (weights[R]-percentile)/(weights[R]-weights[L])
+
+            return LP*values[L] + (1-LP)*values[R]
         else:
-            return values[int(i)] * (1-(i-int(i))) + values[int(i)+1] * (i-int(i))
+            i = percentile*(len(values)-1)
+            I = int(i)
+            
+            if i == I:
+                return values[I]
+            else:
+                w = (i-I)
+                return (1-w)*values[I] + w*values[I+1]
 
-    values = sorted(values)
+    if sort:
+        if weights:
+            values, weights = zip(*sorted(zip(values,weights)))
+        else:
+            values = sorted(values)
+
+    if weights:
+        if any(not w for w in weights):
+            values = list(compress(values,weights))
+        weight_sum = sum(weights[1:])
+        weights    = [w/weight_sum for w in accumulate(chain([0],compress(weights[1:],weights)))]
 
     if isinstance(percentiles,(float,int)):
-        return _percentile(values, percentiles)
+        return _percentile(values, weights, percentiles)
     else:
-        return tuple([_percentile(values, p) for p in percentiles ])
+        return tuple([_percentile(values, weights, p) for p in percentiles ])
 
 def phi(x: float) -> float:
     'Cumulative distribution function for the standard normal distribution'
     return (1.0 + erf(x / sqrt(2.0))) / 2.0
 
 def mean(sample: Sequence[float]) -> float:
-    #If precision is needed use a true statistics package  
-    return sum(sample)/len(sample)
+    return fmean(sample)
 
 def var(sample: Sequence[float]) -> float:
     n = len(sample)
@@ -59,7 +90,7 @@ def var(sample: Sequence[float]) -> float:
         E_s2 = hypot(*sample)**2
         E_s = sum(sample)
         return (E_s2-E_s*E_s/n)/(n-1)
-    else:
+    else: #pragma: no cover
         #using the corrected two pass algo as recommended by
         #https://cpsc.yale.edu/sites/default/files/files/tr222.pdf
         #I've optimized this as much as I think is possible in python
@@ -72,45 +103,69 @@ def stdev(sample: Sequence[float]) -> float:
 class PointAndInterval(ABC):
 
     @abstractmethod
-    def calculate(self, sample: Sequence[float]) -> Tuple[float, Tuple[float, float]]:
+    def point(self, sample: Sequence[float]) -> float:
         ...
 
-class StandardErrorOfMean(PointAndInterval):
+    @abstractmethod
+    def point_interval(self, sample: Sequence[float]) -> Tuple[float, Tuple[float, float]]:
+        ...
+
+class StdDevCI(PointAndInterval):
+    def point(self,sample: Sequence[float]) -> float:
+        return mean(sample)
+    def point_interval(self, sample: Sequence[float]) -> Tuple[float, Tuple[float, float]]:
+        mu = mean(sample)
+        sd = round(0 if len(sample) == 1 else stdev(sample),5)
+        return (mu, (sd,sd))
+
+class StdErrCI(PointAndInterval):
 
     def __init__(self, z_score:float=1.96) -> None:
-        self._z_score = z_score 
+        self._z_score = z_score
+        self._inner   = StdDevCI()
 
-    def calculate(self, sample: Sequence[float]) -> Tuple[float, Tuple[float, float]]:
-        mu = mean(sample)
-        se = stdev(sample)/(len(sample)**(.5))
-        return (mu, (self._z_score*se,self._z_score*se))
+    def point(self,sample: Sequence[float]) -> float:
+        return self._inner.point(sample)
 
-class BootstrapConfidenceInterval(PointAndInterval):
+    def point_interval(self, sample: Sequence[float]) -> Tuple[float, Tuple[float, float]]:
+        (mu,(sd,sd)) = self._inner.point_interval(sample)
+        sqrtn = len(sample)**.5
+        se    = round(sd/sqrtn,5)
+        ci    = self._z_score*se
+        return (mu, (ci,ci))
+
+class BootstrapCI(PointAndInterval):
 
     def __init__(self, confidence:float, statistic:Callable[[Sequence[float]], float]) -> None:
+        PackageChecker.scipy('BootstrapConfidenceInterval')
         self._conf = confidence
         self._stat = statistic
+        self._args = dict(method='basic', vectorized=False, n_resamples=1000, random_state=1)
 
-    def calculate(self, sample: Sequence[float]) -> Tuple[float, Tuple[float, float]]:
-        rng = CobaRandom(1)
-        n = len(sample)
+    def point(self,sample: Sequence[float]) -> float:
+        return self._stat(sample)
 
-        sample_stats = [ self._stat([sample[i] for i in rng.randints(n, 0, n-1)]) for _ in range(50) ]
+    def point_interval(self, sample: Sequence[float]) -> Tuple[float, Tuple[float, float]]:
+        from scipy.stats import bootstrap
 
-        lower_conf = (1-self._conf)/2
-        upper_conf = (1+self._conf)/2
+        p   = self._stat(sample)
 
-        point_stat = self._stat(sample)
-        lower_stat,upper_stat = percentile(sample_stats,[lower_conf,upper_conf])
+        if len(sample) < 3:
+            l,h = p,p
+        else:
+            l,h = bootstrap([sample], self._stat, **self._args).confidence_interval
 
-        return (point_stat, (point_stat-lower_stat,upper_stat-point_stat))
+        return (p, (p-l,h-p))
 
-class BinomialConfidenceInterval(PointAndInterval):
+class BinomialCI(PointAndInterval):
 
     def __init__(self, method:Literal['wilson', 'clopper-pearson']):
         self._method = method
 
-    def calculate(self, sample: Sequence[float]) -> Tuple[float, Tuple[float, float]]:
+    def point(self,sample: Sequence[float]) -> float:
+        return sum(sample)/len(sample)
+
+    def point_interval(self, sample: Sequence[float]) -> Tuple[float, Tuple[float, float]]:
         if set(sample) - set([0,1]):
             raise CobaException("A binomial confidence interval can only be calculated on values of 0 and 1.")
 
@@ -123,7 +178,7 @@ class BinomialConfidenceInterval(PointAndInterval):
             #https://www.itl.nist.gov/div898/handbook/prc/section2/prc241.htm
             interval_num = z_975*((p_hat*(1-p_hat))/n + Q/(2*n))**(.5)
             location_num = (p_hat+Q)
-            
+
             interval_den = (1+2*Q)
             location_den = (1+2*Q)
 
@@ -131,10 +186,11 @@ class BinomialConfidenceInterval(PointAndInterval):
             location = location_num/location_den
 
             return (p_hat, (p_hat-(location-interval), (location+interval)-p_hat))
-        
+
         else:
-            PackageChecker.sklearn("BinomialConfidenceInterval")
+            PackageChecker.scipy("BinomialConfidenceInterval")
             from scipy.stats import beta
+
             lo = beta.ppf(.05/2, sum(sample), len(sample) - sum(sample) + 1)
             hi = beta.ppf(1-.05/2, sum(sample) + 1, len(sample) - sum(sample))
             p_hat = sum(sample)/len(sample)
@@ -144,7 +200,7 @@ class BinomialConfidenceInterval(PointAndInterval):
 
             return (p_hat, (p_hat-lo,hi-p_hat))
 
-class OnlineVariance():
+class OnlineVariance:
     """Calculate sample variance in an online fashion.
 
     Remarks:
@@ -185,7 +241,7 @@ class OnlineVariance():
         if count > 1:
             self._variance = M2 / (count - 1)
 
-class OnlineMean():
+class OnlineMean:
     """Calculate mean in an online fashion."""
 
     def __init__(self):
