@@ -1,9 +1,9 @@
 import warnings
 import importlib
 
-from itertools import chain
+from itertools import chain, islice
 from collections import defaultdict
-from typing import TypeVar, Iterable, Tuple, Sequence
+from typing import TypeVar, Iterable, Tuple, Union, Sequence
 
 from coba.exceptions import CobaExit
 
@@ -22,15 +22,8 @@ class PackageChecker:
 
         Args:
             caller_name: The name of the caller that requires matplotlib.
-
-        Remarks:
-            This pattern borrows heavily from sklearn. As of 6/20/2020 sklearn code could be found
-            at https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/utils/__init__.py
         """
-        try:
-            importlib.import_module('matplotlib.pyplot')
-        except ImportError:
-            PackageChecker._handle_import_error(caller_name, "matplotlib")
+        PackageChecker._handle_import_error(caller_name, "matplotlib.pyplot", "matplotlib")
 
     @staticmethod
     def vowpalwabbit(caller_name: str) -> None:
@@ -40,15 +33,9 @@ class PackageChecker:
 
         Args:
             caller_name: The name of the caller that requires vowpalwabbit.
-
-        Remarks:
-            This pattern was inspired by sklearn (see `PackageChecker.matplotlib` for more information).
         """
 
-        try:
-            importlib.import_module('vowpalwabbit')
-        except ImportError as e:
-            PackageChecker._handle_import_error(caller_name, "vowpalwabbit")
+        PackageChecker._handle_import_error(caller_name, "vowpalwabbit")
 
     @staticmethod
     def pandas(caller_name: str) -> None:
@@ -58,14 +45,8 @@ class PackageChecker:
 
         Args:
             caller_name: The name of the caller that requires pandas.
-
-        Remarks:
-            This pattern was inspired by sklearn (see `PackageChecker.matplotlib` for more information).
         """
-        try:
-            importlib.import_module('pandas')
-        except ImportError:
-            PackageChecker._handle_import_error(caller_name, "pandas")
+        PackageChecker._handle_import_error(caller_name, "pandas")
 
     @staticmethod
     def numpy(caller_name: str) -> None:
@@ -75,14 +56,19 @@ class PackageChecker:
 
         Args:
             caller_name: The name of the caller that requires numpy.
-
-        Remarks:
-            This pattern was inspired by sklearn (see `PackageChecker.matplotlib` for more information).
         """
-        try:
-            importlib.import_module('numpy')
-        except ImportError:
-            PackageChecker._handle_import_error(caller_name, "numpy")
+        PackageChecker._handle_import_error(caller_name, "numpy")
+
+    @staticmethod
+    def scipy(caller_name: str) -> None:
+        """Raise ImportError with detailed error message if scipy is not installed.
+
+        Functionality requiring scipy should call this helper and then lazily import.
+
+        Args:
+            caller_name: The name of the caller that requires scipy.
+        """
+        PackageChecker._handle_import_error(caller_name, "scipy")
 
     @staticmethod
     def sklearn(caller_name: str) -> None:
@@ -92,17 +78,31 @@ class PackageChecker:
 
         Args:
             caller_name: The name of the caller that requires sklearn.
+        """
+        PackageChecker._handle_import_error(caller_name, "sklearn", "scikit-learn")
 
+    @staticmethod
+    def torch(caller_name: str) -> None:
+        """Raise ImportError with detailed error message if torch is not installed.
+
+        Functionality requiring torch should call this helper and then lazily import.
+
+        Args:
+            caller_name: The name of the caller that requires torch.
+        """
+        PackageChecker._handle_import_error(caller_name, "torch")
+
+    def _handle_import_error(caller_name:str, module_name:str, pkg_name:str = None):
+        """
         Remarks:
-            This pattern was inspired by sklearn (see `PackageChecker.matplotlib` for more information).
+            This pattern borrows heavily from sklearn. As of 6/20/2020 sklearn code could be found
+            at https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/utils/__init__.py
         """
         try:
-            importlib.import_module('sklearn')
+            importlib.import_module(module_name)
         except ImportError:
-            PackageChecker._handle_import_error(caller_name, "scikit-learn")
-
-    def _handle_import_error(caller_name:str, pkg_name:str):
-        coba_exit(f"ERROR: {caller_name} requires the {pkg_name} package. You can install this package via `pip install {pkg_name}`.")
+            pkg_name = pkg_name or module_name
+            coba_exit(f"ERROR: {caller_name} requires the {pkg_name} package. You can install this package via `pip install {pkg_name}`.")
 
 class KeyDefaultDict(defaultdict):
     def __missing__(self, key):
@@ -114,10 +114,11 @@ class KeyDefaultDict(defaultdict):
             return value
 
 _T = TypeVar("_T")
-def peek_first(items: Iterable[_T]) -> Tuple[_T, Iterable[_T]]:
+def peek_first(items: Iterable[_T], n:int=1) -> Tuple[Union[_T,Sequence[_T]], Iterable[_T]]:
     items = iter(items)
-    try:
-        first = next(items)
-        return first, chain([first],items)
-    except StopIteration:
-        return None, []
+    first = list(islice(items,n))
+
+    items = [] if not first else chain(first,items)
+    first = None if not first else first[0] if n==1 else first
+
+    return first, items
